import pytest
from typing import Any, Optional

from pydantic import Field

from vellum.workflows.errors import WorkflowErrorCode
from vellum.workflows.exceptions import WorkflowInitializationException
from vellum.workflows.inputs import BaseInputs


def test_base_inputs_happy_path():
    # GIVEN some input class with required and optional fields
    class TestInputs(BaseInputs):
        required_string: str
        required_int: int
        optional_string: Optional[str]

    # WHEN we assign the inputs some valid values
    inputs = TestInputs(required_string="hello", required_int=42, optional_string=None)

    # THEN the inputs should have the correct values
    assert inputs.required_string == "hello"
    assert inputs.required_int == 42
    assert inputs.optional_string is None


def test_base_inputs_explicit_none():
    """
    Test that None can be explicitly set as a value, distinguishing it from not providing a value.
    """

    # GIVEN some input class with optional fields and fields with defaults
    class TestInputs(BaseInputs):
        optional_string: Optional[str]
        optional_string_with_default: Optional[str] = None

    # WHEN we explicitly pass None for optional fields
    inputs = TestInputs(optional_string=None)

    assert inputs.optional_string is None
    assert inputs.optional_string_with_default is None


@pytest.mark.parametrize("field_type", [str, Optional[str]])
def test_base_inputs_explicit_none_should_raise_on_fields_without_defaults(field_type):
    """
    Test that None cannot be explicitly set as a value for required fields.
    """

    class TestInputs(BaseInputs):
        required_string: field_type  # type: ignore[valid-type]

    with pytest.raises(WorkflowInitializationException) as exc_info:
        TestInputs()  # type: ignore[call-arg]

    assert exc_info.value.code == WorkflowErrorCode.INVALID_INPUTS
    assert "Required input variables required_string should have defined value" == str(exc_info.value)


def test_base_inputs_explicit_none_should_raise_on_required_fields_with_none():
    """
    Test that None cannot be explicitly set as a value for required fields.
    """

    class TestInputs(BaseInputs):
        required_string: str

    with pytest.raises(WorkflowInitializationException) as exc_info:
        TestInputs(required_string=None)  # type: ignore[arg-type]

    assert exc_info.value.code == WorkflowErrorCode.INVALID_INPUTS
    assert "Required input variables required_string should have defined value" == str(exc_info.value)


def test_base_inputs_empty_value():
    # GIVEN some input class with required and optional string fields
    class TestInputs(BaseInputs):
        required_string: str
        optional_string: Optional[str]

    # WHEN we try to omit a required field
    with pytest.raises(WorkflowInitializationException) as exc_info:
        TestInputs(optional_string="ok")  # type: ignore

    # THEN it should raise a NodeException with the correct error message and code
    assert exc_info.value.code == WorkflowErrorCode.INVALID_INPUTS
    assert "Required input variables required_string should have defined value" == str(exc_info.value)


def test_base_inputs_with_default():
    # GIVEN some input class with a field that has a default value
    class TestInputs(BaseInputs):
        string_with_default: str = "default_value"

    # WHEN we create an instance without providing the field
    inputs = TestInputs()

    # THEN it should use the default value
    assert inputs.string_with_default == "default_value"


def test_base_inputs_with_field_default_factory_empty_list():
    """
    Test that Field(default_factory=list) works correctly for JSON array inputs.
    This ensures that each instance gets a separate list instance, avoiding mutable default issues.
    """

    # GIVEN some input class with Field(default_factory=list) for JSON array
    class TestInputs(BaseInputs):
        json_array_input: Any = Field(default_factory=list)  # type: ignore[arg-type]
        array_input: list[str] = Field(default_factory=list)  # type: ignore[arg-type]

    # WHEN we create two instances without providing values
    inputs1 = TestInputs()
    inputs2 = TestInputs()

    # THEN they should get separate list instances
    assert inputs1.json_array_input is not inputs2.json_array_input
    assert isinstance(inputs1.json_array_input, list)
    assert isinstance(inputs2.json_array_input, list)

    # WHEN we modify one instance
    inputs1.json_array_input.append("test1")
    inputs2.json_array_input.append("test2")

    # THEN the other instance should be unaffected
    assert inputs1.json_array_input == ["test1"]
    assert inputs2.json_array_input == ["test2"]


def test_base_inputs_with_field_default_factory_non_empty_list():
    """
    Test that Field(default_factory=list) works correctly for non-empty list inputs.
    """

    # GIVEN some input class with Field(default_factory=list) for non-empty list
    class TestInputs(BaseInputs):
        non_empty_list_input: list[str] = Field(default_factory=lambda: ["test1", "test2"])

    # WHEN we create an instance without providing values
    inputs1 = TestInputs()
    inputs2 = TestInputs()

    # THEN it should use the default value
    assert inputs1.non_empty_list_input == ["test1", "test2"]
    assert inputs2.non_empty_list_input == ["test1", "test2"]

    # WHEN we modify one instance
    inputs1.non_empty_list_input.append("test3")
    inputs2.non_empty_list_input.append("test4")

    # THEN the other instance should be unaffected
    assert inputs1.non_empty_list_input == ["test1", "test2", "test3"]
    assert inputs2.non_empty_list_input == ["test1", "test2", "test4"]


def test_base_inputs_with_field_default_factory_empty_dict():
    """
    Test that Field(default_factory=dict) works correctly for JSON dict inputs.
    This ensures that each instance gets a separate dict instance, avoiding mutable default issues.
    """

    # GIVEN some input class with Field(default_factory=dict) for JSON dict
    class TestInputs(BaseInputs):
        json_dict_input: Any = Field(default_factory=dict)  # type: ignore[arg-type]

    # WHEN we create two instances without providing values
    inputs1 = TestInputs()
    inputs2 = TestInputs()

    # THEN they should get separate dict instances
    assert inputs1.json_dict_input is not inputs2.json_dict_input
    assert isinstance(inputs1.json_dict_input, dict)
    assert isinstance(inputs2.json_dict_input, dict)

    # WHEN we modify one instance
    inputs1.json_dict_input["key1"] = "value1"
    inputs2.json_dict_input["key2"] = "value2"

    # THEN the other instance should be unaffected
    assert inputs1.json_dict_input == {"key1": "value1"}
    assert inputs2.json_dict_input == {"key2": "value2"}


def test_base_inputs_with_field_default_factory_non_empty_dict():
    """
    Test that Field(default_factory=dict) works correctly for non-empty dict inputs.
    """

    # GIVEN some input class with Field(default_factory=dict) for non-empty dict
    class TestInputs(BaseInputs):
        non_empty_dict_input: dict[str, str] = Field(default_factory=lambda: {"key1": "value1", "key2": "value2"})

    # WHEN we create an instance without providing values
    inputs1 = TestInputs()
    inputs2 = TestInputs()

    # THEN it should use the default value
    assert inputs1.non_empty_dict_input == {"key1": "value1", "key2": "value2"}
    assert inputs2.non_empty_dict_input == {"key1": "value1", "key2": "value2"}

    # WHEN we modify one instance
    inputs1.non_empty_dict_input["key3"] = "value3"
    inputs2.non_empty_dict_input["key4"] = "value4"

    # THEN the other instance should be unaffected
    assert inputs1.non_empty_dict_input == {"key1": "value1", "key2": "value2", "key3": "value3"}
    assert inputs2.non_empty_dict_input == {"key1": "value1", "key2": "value2", "key4": "value4"}


def test_base_inputs__supports_inherited_inputs():
    # GIVEN an inputs class
    class TopInputs(BaseInputs):
        first: str

    # WHEN we inherit from the base inputs class
    class BottomInputs(TopInputs):
        second: int

    # THEN both references should be available
    assert BottomInputs.first.name == "first"
    assert BottomInputs.second.name == "second"
    assert len([ref for ref in BottomInputs]) == 2


def test_base_inputs__iterating_over_descriptor_is_finite():
    """
    Tests that iterating over a descriptor (like Inputs.image_url) creates a finite generator.
    This was the bug reported in Slack where AB sometimes generates code that iterates over descriptors.
    """

    class Inputs(BaseInputs):
        image_url: str

    iteration_count = 0
    try:
        for idx, x in enumerate(Inputs.image_url):  # type: ignore[arg-type,var-annotated]
            iteration_count += 1
            if idx > 100:
                break
    except (TypeError, IndexError):
        pass

    assert iteration_count < 100



# This file was auto-generated by Fern from our API Definition.

import typing

import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel


class PromptParameters(UniversalBaseModel):
    stop: typing.Optional[typing.List[str]] = None
    temperature: typing.Optional[float] = None
    max_tokens: typing.Optional[int] = None
    top_p: typing.Optional[float] = None
    top_k: typing.Optional[int] = None
    frequency_penalty: typing.Optional[float] = None
    presence_penalty: typing.Optional[float] = None
    logit_bias: typing.Optional[typing.Dict[str, typing.Optional[float]]] = None
    custom_parameters: typing.Optional[typing.Dict[str, typing.Optional[typing.Any]]] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



# This file was auto-generated by Fern from our API Definition.

import typing

from .folder_entity_dataset import FolderEntityDataset
from .folder_entity_document_index import FolderEntityDocumentIndex
from .folder_entity_folder import FolderEntityFolder
from .folder_entity_prompt_sandbox import FolderEntityPromptSandbox
from .folder_entity_test_suite import FolderEntityTestSuite
from .folder_entity_workflow_sandbox import FolderEntityWorkflowSandbox

FolderEntity = typing.Union[
    FolderEntityFolder,
    FolderEntityPromptSandbox,
    FolderEntityWorkflowSandbox,
    FolderEntityDocumentIndex,
    FolderEntityTestSuite,
    FolderEntityDataset,
]



from importlib import metadata
import io
import json
import os
import sys
import tarfile
from uuid import UUID
from typing import List, Optional

from dotenv import load_dotenv
from pydantic import ValidationError

from vellum.client.core.api_error import ApiError
from vellum.client.types import WorkflowPushDeploymentConfigRequest
from vellum.resources.workflows.client import OMIT
from vellum.workflows.vellum_client import create_vellum_client
from vellum_cli.config import DEFAULT_WORKSPACE_CONFIG, WorkflowConfig, WorkflowDeploymentConfig, load_vellum_cli_config
from vellum_cli.logger import handle_cli_error, load_cli_logger
from vellum_ee.workflows.display.nodes.utils import to_kebab_case
from vellum_ee.workflows.display.workflows.base_workflow_display import BaseWorkflowDisplay


def push_command(
    module: Optional[str] = None,
    workflow_sandbox_id: Optional[str] = None,
    deploy: Optional[bool] = None,
    deployment_label: Optional[str] = None,
    deployment_name: Optional[str] = None,
    deployment_description: Optional[str] = None,
    release_tags: Optional[List[str]] = None,
    release_description: Optional[str] = None,
    dry_run: Optional[bool] = None,
    strict: Optional[bool] = None,
    workspace: Optional[str] = None,
) -> None:
    load_dotenv(dotenv_path=os.path.join(os.getcwd(), ".env"))
    logger = load_cli_logger()
    config = load_vellum_cli_config()

    workflow_configs = (
        [
            w
            for w in config.workflows
            if (module and w.module == module) or (workflow_sandbox_id and w.workflow_sandbox_id == workflow_sandbox_id)
        ]
        if module or workflow_sandbox_id
        else config.workflows
    )

    if len(workflow_configs) > 1:
        if workspace:
            workflow_configs = [w for w in workflow_configs if w.workspace == workspace]
        else:
            workflow_configs = [w for w in workflow_configs if w.workspace == DEFAULT_WORKSPACE_CONFIG.name]

    if len(workflow_configs) == 0:
        if module and module_exists(module):
            new_config = WorkflowConfig(
                module=module,
                workflow_sandbox_id=workflow_sandbox_id,
                workspace=workspace or DEFAULT_WORKSPACE_CONFIG.name,
            )
            config.workflows.append(new_config)
            workflow_configs = [new_config]
        elif not module:
            handle_cli_error(
                logger,
                title="No workflows found in project to push",
                message="No workflow configurations were found in vellum.lock.json.",
                suggestion="If you pulled a workflow with a specific sandbox ID"
                "(e.g., vellum workflows pull --workflow-sandbox-id=<id>), "
                "you should push without specifying a different sandbox ID:\n\n"
                "  vellum workflows push\n\n"
                "Or, if you want to push to a specific module:\n\n"
                "  vellum workflows push --module=<module-name>",
            )
            return
        else:
            raise ValueError(f"No workflow config for '{module}' found in project to push.")

    if len(workflow_configs) > 1:
        raise ValueError(
            "Multiple workflows found. Please specify a single workflow to push. "
            f"Found: {', '.join([w.module for w in workflow_configs])}"
        )

    workflow_config = workflow_configs[0]

    logger.info(f"Loading workflow from {workflow_config.module}")
    resolved_workspace = workspace or workflow_config.workspace or DEFAULT_WORKSPACE_CONFIG.name
    workspace_config = next((w for w in config.workspaces if w.name == resolved_workspace), None)

    if workspace_config is None:
        if resolved_workspace == DEFAULT_WORKSPACE_CONFIG.name:
            workspace_config = DEFAULT_WORKSPACE_CONFIG
        else:
            available_workspaces = sorted(set([w.name for w in config.workspaces] + [DEFAULT_WORKSPACE_CONFIG.name]))
            handle_cli_error(
                logger,
                title=f"Workspace '{resolved_workspace}' not found in config",
                message=f"Available workspaces: {', '.join(available_workspaces)}",
            )
            return

    api_key = os.getenv(workspace_config.api_key)
    if not api_key:
        raise ValueError(f"No API key value found in environment for workspace '{workspace_config.name}'.")

    if workspace_config.name != workflow_config.workspace:
        # We are pushing to a new workspace, so we need a new workflow config
        workflow_config = WorkflowConfig(
            module=workflow_config.module,
            workspace=workspace_config.name,
        )
        config.workflows.append(workflow_config)

    if (
        workflow_sandbox_id
        and workflow_config.workflow_sandbox_id
        and workflow_config.workflow_sandbox_id != workflow_sandbox_id
    ):
        logger.warning(
            f"Workflow sandbox id '{workflow_sandbox_id}' is already associated with '{workflow_config.module}'. "
            f"Continuing with the provided workflow sandbox id '{workflow_sandbox_id}'."
        )

    client = create_vellum_client(
        api_key=api_key,
        api_url=workspace_config.api_url,
    )
    sys.path.insert(0, os.getcwd())

    # Remove this once we could serialize using the artifact in Vembda
    # https://app.shortcut.com/vellum/story/5585
    serialization_result = BaseWorkflowDisplay.serialize_module(
        workflow_config.module,
        client=client,
        dry_run=dry_run or False,
    )
    exec_config = serialization_result.exec_config

    container_tag = workflow_config.container_image_tag
    if workflow_config.container_image_name and not workflow_config.container_image_tag:
        container_tag = "latest"

    exec_config["runner_config"] = {
        "sdk_version": metadata.version("vellum-ai"),  # Deprecated in favor of codegen_version
        "codegen_version": metadata.version("vellum-ai"),
        "container_image_tag": container_tag,
        "container_image_name": workflow_config.container_image_name,
    }

    deployment_config: WorkflowPushDeploymentConfigRequest = OMIT
    deployment_config_serialized: str = OMIT
    if deploy:
        cli_deployment_config = (
            workflow_config.deployments[0] if workflow_config.deployments else WorkflowDeploymentConfig()
        )

        try:
            module_name = workflow_config.module.split(".")[-1]
            deployment_config = WorkflowPushDeploymentConfigRequest(
                label=deployment_label or cli_deployment_config.label,
                name=deployment_name or cli_deployment_config.name or to_kebab_case(module_name),
                description=deployment_description or cli_deployment_config.description,
                release_tags=release_tags or cli_deployment_config.release_tags,
                release_description=release_description or cli_deployment_config.release_description,
            )
        except ValidationError as e:
            for error in e.errors():
                if "release_tags" in str(error.get("loc", [])):
                    handle_cli_error(
                        logger,
                        title="Invalid release tag format",
                        message="Release tags must be provided as separate arguments. "
                        "Use: --release-tag tag1 --release-tag tag2",
                    )
                    return

            # Re-raise if it's not a release_tags validation error
            raise e

        # We should check with fern if we could auto-serialize typed fields for us
        # https://app.shortcut.com/vellum/story/5568
        deployment_config_serialized = json.dumps({k: v for k, v in deployment_config.dict().items() if v is not None})

    artifact = io.BytesIO()
    with tarfile.open(fileobj=artifact, mode="w:gz") as tar:
        module_dir = workflow_config.module.replace(".", os.path.sep)
        for root, _, files in os.walk(module_dir):
            for filename in files:
                if not filename.endswith(".py"):
                    continue

                file_path = os.path.join(root, filename)
                # Get path relative to module_dir for tar archive
                relative_path = os.path.relpath(file_path, module_dir)
                content_bytes = open(file_path, "rb").read()
                file_buffer = io.BytesIO(content_bytes)

                tarinfo = tarfile.TarInfo(name=relative_path)
                tarinfo.size = len(content_bytes)

                tar.addfile(tarinfo, file_buffer)

    artifact.seek(0)
    artifact.name = f"{workflow_config.module.replace('.', '__')}.tar.gz"

    provided_id = workflow_config.workflow_sandbox_id or workflow_sandbox_id

    try:
        response = client.workflows.push(
            # Remove this once we could serialize using the artifact in Vembda
            # https://app.shortcut.com/vellum/story/5585
            exec_config=json.dumps(exec_config),
            workflow_sandbox_id=provided_id,
            artifact=artifact,
            # We should check with fern if we could auto-serialize typed object fields for us
            # https://app.shortcut.com/vellum/story/5568
            deployment_config=deployment_config_serialized,  # type: ignore[arg-type]
            dry_run=dry_run,
            strict=strict,
        )
    except ApiError as e:
        if e.status_code == 404:
            if provided_id:
                handle_cli_error(
                    logger,
                    title="Workflow Sandbox not found",
                    message=f"Could not find Workflow Sandbox with ID '{provided_id}' "
                    f"in workspace '{resolved_workspace}'.",
                )
            else:
                error_detail = e.body.get("detail") if isinstance(e.body, dict) else None
                default_message = "The `/workflows/push` endpoint failed with a 404 response."
                handle_cli_error(
                    logger,
                    title="Workflow Sandbox not found",
                    message=error_detail or default_message,
                )
            return

        if e.status_code == 400 and isinstance(e.body, dict) and "diffs" in e.body:
            diffs: dict = e.body["diffs"]
            generated_only = diffs.get("generated_only", [])
            generated_only_str = (
                "\n".join(
                    ["Files that were generated but not found in the original project:"]
                    + [f"- {file}" for file in generated_only]
                )
                if generated_only
                else ""
            )

            original_only = diffs.get("original_only", [])
            original_only_str = (
                "\n".join(
                    ["Files that were found in the original project but not generated:"]
                    + [f"- {file}" for file in original_only]
                )
                if original_only
                else ""
            )

            modified = diffs.get("modified", {})
            modified_str = (
                "\n\n".join(
                    ["Files that were different between the original project and the generated artifact:"]
                    + ["\n".join(line.strip() for line in lines) for lines in modified.values()]
                )
                if modified
                else ""
            )

            reported_diffs = f"""\
{e.body.get("detail")}

{generated_only_str}

{original_only_str}

{modified_str}
"""
            logger.error(reported_diffs)
            return

        if e.status_code == 400 and isinstance(e.body, dict) and "detail" in e.body:
            handle_cli_error(logger, title="API request to /workflows/push failed.", message=e.body["detail"])
            return

        raise e

    if dry_run:
        error_messages = [error.message for error in serialization_result.errors]
        error_message = "\n".join(error_messages) if error_messages else "No errors found."
        logger.info(
            f"""\
# Workflow Push Report

## Errors
{error_message}

## Proposed Diffs
{json.dumps(response.proposed_diffs, indent=2)}
"""
        )  # type: ignore[attr-defined]

        error_list = serialization_result.errors
        has_errors = len(error_list) > 0
        has_diffs = response.proposed_diffs is not None and response.proposed_diffs

        if has_errors or has_diffs:
            exit(1)
    else:
        default_api_url = client._client_wrapper._environment.default
        base_url = default_api_url.split("/v1")[0].replace("//api.", "//app.")
        logger.info(
            f"""Successfully pushed {workflow_config.module} to Vellum!
Visit at: {base_url}/workflow-sandboxes/{response.workflow_sandbox_id}"""
        )

    if not workflow_config.workflow_sandbox_id:
        workflow_config.workflow_sandbox_id = response.workflow_sandbox_id

    if not workflow_config.deployments and response.workflow_deployment_id:
        stored_deployment_config = WorkflowDeploymentConfig(
            id=UUID(response.workflow_deployment_id),
            label=deployment_config.label if deploy else None,
            name=deployment_config.name if deploy else None,
            description=deployment_config.description if deploy else None,
            release_tags=deployment_config.release_tags if deploy else None,
            release_description=getattr(deployment_config, "release_description", None) if deploy else None,
        )
        workflow_config.deployments.append(stored_deployment_config)

    config.save()
    logger.info("Updated vellum.lock.json file.")


def module_exists(module_name: str) -> bool:
    module_path = os.path.join(os.getcwd(), *module_name.split("."))
    return os.path.exists(module_path) and os.path.isdir(module_path)



# This file was auto-generated by Fern from our API Definition.

import typing

NewMemberJoinBehaviorEnum = typing.Union[
    typing.Literal["AUTO_ACCEPT_FROM_SHARED_DOMAIN", "ALLOW_REQUESTS_FROM_SHARED_DOMAIN", "REQUIRE_EXPLICIT_INVITE"],
    typing.Any,
]



from deepdiff import DeepDiff

from vellum_ee.workflows.display.workflows.get_vellum_workflow_display_class import get_workflow_display

from tests.workflows.basic_merge_node.await_all_workflow import AwaitAllPassingWorkflow


def test_serialize_workflow__await_all():
    # GIVEN a Workflow that uses an await all merge node
    # WHEN we serialize it
    workflow_display = get_workflow_display(workflow_class=AwaitAllPassingWorkflow)
    serialized_workflow: dict = workflow_display.serialize()

    # THEN we should get a serialized representation of the Workflow
    assert serialized_workflow.keys() == {
        "workflow_raw_data",
        "input_variables",
        "state_variables",
        "output_variables",
    }

    # AND its input variables should be what we expect
    input_variables = serialized_workflow["input_variables"]
    assert len(input_variables) == 0

    # AND its output variables should be what we expect
    output_variables = serialized_workflow["output_variables"]
    assert len(output_variables) == 1
    assert not DeepDiff(
        [
            {"id": "959ba00d-d30b-402e-8676-76efc4c3d2ae", "key": "value", "type": "STRING"},
        ],
        output_variables,
        ignore_order=True,
    )

    # AND its raw data should be what we expect
    workflow_raw_data = serialized_workflow["workflow_raw_data"]
    assert len(workflow_raw_data["edges"]) == 6
    assert len(workflow_raw_data["nodes"]) == 6

    # AND each node should be serialized correctly
    entrypoint_node = next(node for node in workflow_raw_data["nodes"] if node["type"] == "ENTRYPOINT")
    assert entrypoint_node == {
        "id": "dc8aecd0-49ba-4464-a45f-29d3bfd686e4",
        "type": "ENTRYPOINT",
        "base": None,
        "definition": None,
        "inputs": [],
        "data": {
            "label": "Entrypoint Node",
            "source_handle_id": "017d40f5-8326-4e42-a409-b08995defaa8",
        },
        "display_data": {
            "position": {"x": 0.0, "y": -50.0},
        },
    }

    passthrough_nodes = [node for node in workflow_raw_data["nodes"] if node["type"] == "GENERIC"]
    assert len(passthrough_nodes) == 3

    merge_node = next(node for node in workflow_raw_data["nodes"] if node["type"] == "MERGE")
    assert not DeepDiff(
        {
            "id": "f07c263c-65a3-4b58-83c1-f4a29123f167",
            "type": "MERGE",
            "inputs": [],
            "data": {
                "label": "Await All Merge Node",
                "merge_strategy": "AWAIT_ALL",
                "target_handles": [
                    {"id": "4441c835-7d16-4c43-8599-e948b57eaab1"},
                    {"id": "6da3e50a-8c6d-4de1-8ee9-da26f7c9552f"},
                ],
                "source_handle_id": "da1bdfe9-8e99-4d06-842f-a76af95a713a",
            },
            "display_data": {"position": {"x": 400.0, "y": -50.0}},
            "base": {
                "module": ["vellum", "workflows", "nodes", "displayable", "merge_node", "node"],
                "name": "MergeNode",
            },
            "definition": {
                "module": ["tests", "workflows", "basic_merge_node", "await_all_workflow"],
                "name": "AwaitAllMergeNode",
            },
            "trigger": {
                "id": "1c68b622-cc93-4678-a0c4-89f06b6cad1f",
                "merge_behavior": "AWAIT_ALL",
            },
            "ports": [{"id": "da1bdfe9-8e99-4d06-842f-a76af95a713a", "name": "default", "type": "DEFAULT"}],
        },
        merge_node,
        ignore_order_func=lambda x: x.path() == "root['data']['target_handles']",
    )

    final_output_node = next(node for node in workflow_raw_data["nodes"] if node["type"] == "TERMINAL")
    assert final_output_node == {
        "id": "8187ce10-62b7-4a2c-8c0f-297387915467",
        "type": "TERMINAL",
        "data": {
            "label": "Final Output",
            "name": "value",
            "target_handle_id": "ff55701c-16d3-4348-a633-6a298e71377d",
            "output_id": "959ba00d-d30b-402e-8676-76efc4c3d2ae",
            "output_type": "STRING",
            "node_input_id": "fed9d343-6504-460c-968b-d3f9658193d0",
        },
        "base": {
            "module": [
                "vellum",
                "workflows",
                "nodes",
                "displayable",
                "final_output_node",
                "node",
            ],
            "name": "FinalOutputNode",
        },
        "definition": None,
        "inputs": [
            {
                "id": "fed9d343-6504-460c-968b-d3f9658193d0",
                "key": "node_input",
                "value": {
                    "rules": [
                        {
                            "type": "NODE_OUTPUT",
                            "data": {
                                "node_id": "6c34a839-552b-436d-ba6a-501f663883c8",
                                "output_id": "f6148c39-9ec1-4b10-9d24-d19bec9eeea8",
                            },
                        }
                    ],
                    "combinator": "OR",
                },
            }
        ],
        "display_data": {"position": {"x": 800.0, "y": -50.0}},
    }

    # AND each edge should be serialized correctly
    serialized_edges = workflow_raw_data["edges"]
    assert not DeepDiff(
        [
            {
                "id": "ea747507-68d5-4c72-8d00-a464ff7a55f1",
                "source_node_id": "dc8aecd0-49ba-4464-a45f-29d3bfd686e4",
                "source_handle_id": "017d40f5-8326-4e42-a409-b08995defaa8",
                "target_node_id": "0871708d-8f05-4bc8-b3fb-a8624dae51de",
                "target_handle_id": "01543a95-6bcb-46e2-a16c-570008972b88",
                "type": "DEFAULT",
            },
            {
                "id": "2c28f06b-4f4a-4531-8cf9-860089ce1cc2",
                "source_node_id": "dc8aecd0-49ba-4464-a45f-29d3bfd686e4",
                "source_handle_id": "017d40f5-8326-4e42-a409-b08995defaa8",
                "target_node_id": "0306d2a2-8e2a-49d1-bc4d-4026fbd98c4c",
                "target_handle_id": "5b10f22e-ae81-4ca5-a682-12d3c3ec73c1",
                "type": "DEFAULT",
            },
            {
                "id": "3870f290-8da7-43a8-b875-60510c060380",
                "source_node_id": "0306d2a2-8e2a-49d1-bc4d-4026fbd98c4c",
                "source_handle_id": "4e76416e-4edf-4a4a-9202-bf3ee35fd911",
                "target_node_id": "f07c263c-65a3-4b58-83c1-f4a29123f167",
                "target_handle_id": "4441c835-7d16-4c43-8599-e948b57eaab1",
                "type": "DEFAULT",
            },
            {
                "id": "e89010d4-1f5a-4db9-8b67-1fa68960b142",
                "source_node_id": "0871708d-8f05-4bc8-b3fb-a8624dae51de",
                "source_handle_id": "c11fe2da-9ccd-4434-9330-5e3760abe849",
                "target_node_id": "f07c263c-65a3-4b58-83c1-f4a29123f167",
                "target_handle_id": "6da3e50a-8c6d-4de1-8ee9-da26f7c9552f",
                "type": "DEFAULT",
            },
            {
                "id": "5344fbf3-6a71-4890-93a6-c6a2c5a4e50c",
                "source_node_id": "f07c263c-65a3-4b58-83c1-f4a29123f167",
                "source_handle_id": "da1bdfe9-8e99-4d06-842f-a76af95a713a",
                "target_node_id": "6c34a839-552b-436d-ba6a-501f663883c8",
                "target_handle_id": "413c3699-a37b-49cd-93fd-fe6d041a0119",
                "type": "DEFAULT",
            },
            {
                "id": "3d031c93-09b1-4937-9f98-c30a7ba6823d",
                "source_node_id": "6c34a839-552b-436d-ba6a-501f663883c8",
                "source_handle_id": "48beca40-00f3-4273-b2ea-5e1ac6494839",
                "target_node_id": "8187ce10-62b7-4a2c-8c0f-297387915467",
                "target_handle_id": "ff55701c-16d3-4348-a633-6a298e71377d",
                "type": "DEFAULT",
            },
        ],
        serialized_edges,
        ignore_order=True,
    )

    # AND the display data should be what we expect
    display_data = workflow_raw_data["display_data"]
    assert display_data == {
        "viewport": {
            "x": 0.0,
            "y": 0.0,
            "zoom": 1.0,
        }
    }

    # AND the definition should be what we expect
    definition = workflow_raw_data["definition"]
    assert definition == {
        "name": "AwaitAllPassingWorkflow",
        "module": [
            "tests",
            "workflows",
            "basic_merge_node",
            "await_all_workflow",
        ],
    }



import types
from typing import TYPE_CHECKING, Generic, Type, TypeVar

from vellum.workflows.types.generics import NodeType
from vellum_ee.workflows.display.utils.registry import get_from_node_display_registry

if TYPE_CHECKING:
    from vellum_ee.workflows.display.nodes.base_node_display import BaseNodeDisplay


def get_node_display_class(node_class: Type[NodeType]) -> Type["BaseNodeDisplay"]:
    node_display_class = get_from_node_display_registry(node_class)
    if node_display_class:
        return node_display_class

    base_node_display_class = get_node_display_class(node_class.__bases__[0])

    # mypy gets upset at dynamic TypeVar's, but it's technically allowed by python
    _NodeClassType = TypeVar(f"_{node_class.__name__}Type", bound=node_class)  # type: ignore[misc]
    # `base_node_display_class` is always a Generic class, so it's safe to index into it
    NodeDisplayBaseClass = base_node_display_class[_NodeClassType]  # type: ignore[index]

    NodeDisplayClass = types.new_class(
        f"{node_class.__name__}Display",
        bases=(NodeDisplayBaseClass, Generic[_NodeClassType]),
    )

    return NodeDisplayClass



from typing import TYPE_CHECKING, Iterator, List, Optional, Set, Type, Union

from orderly_set import OrderedSet

from vellum.workflows.edges.edge import Edge
from vellum.workflows.edges.trigger_edge import TriggerEdge
from vellum.workflows.types.generics import NodeType

if TYPE_CHECKING:
    from vellum.workflows.nodes.bases.base import BaseNode
    from vellum.workflows.ports.port import Port
    from vellum.workflows.triggers.base import BaseTrigger


class NoPortsNode:
    """Wrapper for nodes that have no ports defined."""

    def __init__(self, node_class: Type["BaseNode"]):
        self.node_class = node_class

    def __repr__(self) -> str:
        return self.node_class.__name__

    def __rshift__(self, other: "GraphTarget") -> "Graph":
        raise ValueError(
            f"Cannot create edges from {self.node_class.__name__} because it has no ports defined. "
            f"Nodes with empty Ports classes cannot be connected to other nodes."
        )


GraphTargetOfSets = Union[
    Set[NodeType],
    Set["Graph"],
    Set["Port"],
    Set[Union[Type["BaseNode"], "Graph", "Port", "NoPortsNode"]],
]

GraphTarget = Union[
    Type["BaseNode"],
    "Port",
    "Graph",
    "NoPortsNode",
    GraphTargetOfSets,
]


class Graph:
    _entrypoints: Set[Union["Port", "NoPortsNode"]]
    _edges: List[Edge]
    _terminals: Set[Union["Port", "NoPortsNode"]]
    _trigger_edges: List[TriggerEdge]

    def __init__(
        self,
        entrypoints: Set[Union["Port", "NoPortsNode"]],
        edges: List[Edge],
        terminals: Set[Union["Port", "NoPortsNode"]],
        trigger_edges: Optional[List[TriggerEdge]] = None,
    ):
        self._edges = edges
        self._entrypoints = entrypoints
        self._terminals = terminals
        self._trigger_edges = trigger_edges or []

    @staticmethod
    def from_port(port: "Port") -> "Graph":
        ports: Set[Union["Port", "NoPortsNode"]] = {port}
        return Graph(entrypoints=ports, edges=[], terminals=ports)

    @staticmethod
    def from_node(node: Type["BaseNode"]) -> "Graph":
        ports = {port for port in node.Ports}
        if not ports:
            no_ports_node = NoPortsNode(node)
            return Graph(entrypoints={no_ports_node}, edges=[], terminals={no_ports_node})
        ports_set: Set[Union["Port", "NoPortsNode"]] = set(ports)
        return Graph(entrypoints=ports_set, edges=[], terminals=ports_set)

    @staticmethod
    def from_set(targets: GraphTargetOfSets) -> "Graph":
        entrypoints = set()
        edges = OrderedSet[Edge]()
        terminals = set()

        for target in targets:
            if isinstance(target, Graph):
                entrypoints.update(target._entrypoints)
                edges.update(target._edges)
                terminals.update(target._terminals)
            elif hasattr(target, "Ports"):
                entrypoints.update({port for port in target.Ports})
                terminals.update({port for port in target.Ports})
            else:
                # target is a Port
                entrypoints.update({target})
                terminals.update({target})

        return Graph(entrypoints=entrypoints, edges=list(edges), terminals=terminals)

    @staticmethod
    def from_edge(edge: Edge) -> "Graph":
        return Graph(entrypoints={edge.from_port}, edges=[edge], terminals={port for port in edge.to_node.Ports})

    @staticmethod
    def from_trigger_edge(edge: TriggerEdge) -> "Graph":
        """
        Create a graph from a single TriggerEdge (Trigger >> Node).

        Args:
            edge: TriggerEdge connecting a trigger to a node

        Returns:
            Graph with the trigger edge and the target node's ports as terminals
        """
        ports = {port for port in edge.to_node.Ports}
        if not ports:
            no_ports_node = NoPortsNode(edge.to_node)
            return Graph(
                entrypoints={no_ports_node},
                edges=[],
                terminals={no_ports_node},
                trigger_edges=[edge],
            )
        return Graph(
            entrypoints=set(ports),
            edges=[],
            terminals=set(ports),
            trigger_edges=[edge],
        )

    @staticmethod
    def from_trigger_edges(edges: List[TriggerEdge]) -> "Graph":
        """
        Create a graph from multiple TriggerEdges (e.g., Trigger >> {NodeA, NodeB}).

        Args:
            edges: List of TriggerEdges

        Returns:
            Graph with all trigger edges and target nodes' ports as entrypoints/terminals
        """
        entrypoints: Set[Union["Port", NoPortsNode]] = set()
        terminals: Set[Union["Port", NoPortsNode]] = set()

        for edge in edges:
            ports = {port for port in edge.to_node.Ports}
            if not ports:
                no_ports_node = NoPortsNode(edge.to_node)
                entrypoints.add(no_ports_node)
                terminals.add(no_ports_node)
            else:
                entrypoints.update(ports)
                terminals.update(ports)

        return Graph(
            entrypoints=entrypoints,
            edges=[],
            terminals=terminals,
            trigger_edges=edges,
        )

    @staticmethod
    def empty() -> "Graph":
        """Create an empty graph with no entrypoints, edges, or terminals."""
        return Graph(entrypoints=set(), edges=[], terminals=set())

    def __rshift__(self, other: GraphTarget) -> "Graph":
        # Check for trigger target (class-level only)
        from vellum.workflows.triggers.base import BaseTrigger

        if isinstance(other, type) and issubclass(other, BaseTrigger):
            raise TypeError(
                f"Cannot create edge targeting trigger {other.__name__}. "
                f"Triggers must be at the start of a graph path, not as targets."
            )

        if not self._edges and not self._entrypoints:
            raise ValueError(
                "Cannot create edges from nodes with empty Ports classes (like TERMINAL/FinalOutputNode). "
                "TERMINAL nodes are designed to be workflow outputs and cannot connect to other nodes."
            )

        if self._terminals and all(isinstance(terminal, NoPortsNode) for terminal in self._terminals):
            terminal_names = [terminal.node_class.__name__ for terminal in self._terminals]
            raise ValueError(
                f"Cannot create edges from graph because all terminal nodes have no ports defined: "
                f"{', '.join(terminal_names)}. Nodes with empty Ports classes cannot be connected to other nodes."
            )

        if isinstance(other, set):
            new_terminals = set()
            for elem in other:
                for final_output_node in self._terminals:
                    if isinstance(final_output_node, NoPortsNode):
                        continue
                    if isinstance(elem, Graph):
                        midgraph = final_output_node >> set(elem.entrypoints)
                        self._extend_edges(midgraph.edges)
                        self._extend_edges(elem.edges)
                        for other_terminal in elem._terminals:
                            new_terminals.add(other_terminal)
                    elif hasattr(elem, "Ports"):
                        midgraph = final_output_node >> elem
                        self._extend_edges(midgraph.edges)
                        for other_terminal in elem.Ports:
                            new_terminals.add(other_terminal)
                    else:
                        # elem is a Port
                        midgraph = final_output_node >> elem
                        self._extend_edges(midgraph.edges)
                        new_terminals.add(elem)
            self._terminals = new_terminals
            return self

        if isinstance(other, Graph):
            for final_output_node in self._terminals:
                if isinstance(final_output_node, NoPortsNode):
                    continue
                midgraph = final_output_node >> set(other.entrypoints)
                self._extend_edges(midgraph.edges)
                self._extend_edges(other.edges)
            self._terminals = other._terminals
            return self

        if hasattr(other, "Ports"):
            for final_output_node in self._terminals:
                if isinstance(final_output_node, NoPortsNode):
                    continue
                subgraph = final_output_node >> other
                self._extend_edges(subgraph.edges)
            self._terminals = {port for port in other.Ports}
            return self

        # other is a Port
        for final_output_node in self._terminals:
            if isinstance(final_output_node, NoPortsNode):
                continue
            subgraph = final_output_node >> other
            self._extend_edges(subgraph.edges)
        self._terminals = {other}
        return self

    def __rrshift__(cls, other_cls: GraphTarget) -> "Graph":
        if not isinstance(other_cls, set):
            other_cls = {other_cls}

        return Graph.from_set(other_cls) >> cls

    @property
    def entrypoints(self) -> Iterator[Type["BaseNode"]]:
        return iter(e.node_class for e in self._entrypoints)

    @property
    def edges(self) -> Iterator[Edge]:
        return iter(self._edges)

    @property
    def trigger_edges(self) -> Iterator[TriggerEdge]:
        """Get all trigger edges in this graph."""
        return iter(self._trigger_edges)

    @property
    def triggers(self) -> Iterator[Type["BaseTrigger"]]:
        """Get all unique trigger classes in this graph."""
        seen_triggers = set()
        for trigger_edge in self._trigger_edges:
            if trigger_edge.trigger_class not in seen_triggers:
                seen_triggers.add(trigger_edge.trigger_class)
                yield trigger_edge.trigger_class

    @property
    def nodes(self) -> Iterator[Type["BaseNode"]]:
        nodes = set()

        # Include nodes from trigger edges
        for trigger_edge in self._trigger_edges:
            if trigger_edge.to_node not in nodes:
                nodes.add(trigger_edge.to_node)
                yield trigger_edge.to_node

        if not self._edges:
            for node in self.entrypoints:
                if node not in nodes:
                    nodes.add(node)
                    yield node
            return

        for edge in self._edges:
            if edge.from_port.node_class not in nodes:
                nodes.add(edge.from_port.node_class)
                yield edge.from_port.node_class
            if edge.to_node not in nodes:
                nodes.add(edge.to_node)
                yield edge.to_node

    def _extend_edges(self, edges: Iterator[Edge]) -> None:
        for edge in edges:
            if edge not in self._edges:
                self._edges.append(edge)

    def __str__(self) -> str:
        """
        Return a visual ASCII representation of the graph showing the flow structure.
        """
        if not self._edges and not self._entrypoints:
            return "Graph(empty)"

        if not self._edges:
            if len(self._entrypoints) == 1:
                port = next(iter(self._entrypoints))
                return f"Graph: {self._get_port_name(port)}"
            else:
                port_names = [self._get_port_name(port) for port in self._entrypoints]
                return f"Graph: [{', '.join(port_names)}]"

        return self._build_flow_diagram()

    def _build_flow_diagram(self) -> str:
        """Build a connected flow diagram showing the graph structure."""
        lines = ["Graph:"]

        adjacency: dict[str, list[str]] = {}
        all_nodes = set()

        for edge in self._edges:
            source_node = edge.from_port.node_class.__name__
            target_node = edge.to_node.__name__

            all_nodes.add(source_node)
            all_nodes.add(target_node)

            if source_node not in adjacency:
                adjacency[source_node] = []
            adjacency[source_node].append(target_node)

        target_nodes = set()
        for edges in adjacency.values():
            target_nodes.update(edges)

        root_nodes = []
        for port in self._entrypoints:
            node_name = port.node_class.__name__
            if node_name not in target_nodes:
                root_nodes.append(node_name)

        if not root_nodes and self._entrypoints:
            root_nodes = [next(iter(self._entrypoints)).node_class.__name__]

        visited = set()
        currently_visiting = set()

        def render_node(node: str, prefix: str = "  ", is_last: bool = True, path: Optional[List[str]] = None) -> None:
            if path is None:
                path = []

            if node in currently_visiting:
                lines.append(f"{prefix}{'└─' if is_last else '├─'} {node} ⟲ (loops back)")
                return

            if node in visited:
                lines.append(f"{prefix}{'└─' if is_last else '├─'} {node} → (see above)")
                return

            visited.add(node)
            currently_visiting.add(node)

            lines.append(f"{prefix}{'└─' if is_last else '├─'} {node}")

            if node in adjacency:
                children = adjacency[node]
                for i, child in enumerate(children):
                    child_is_last = i == len(children) - 1
                    next_prefix = prefix + ("   " if is_last else "│  ")
                    render_node(child, next_prefix, child_is_last, path + [node])

            currently_visiting.remove(node)

        for i, root in enumerate(root_nodes):
            is_last_root = i == len(root_nodes) - 1
            render_node(root, "  ", is_last_root)

        return "\n".join(lines)

    def _get_port_name(self, port: Union["Port", "NoPortsNode"]) -> str:
        """Get a readable name for a port."""
        if isinstance(port, NoPortsNode):
            return f"{port.node_class.__name__} (no ports)"
        try:
            if hasattr(port, "node_class") and hasattr(port.node_class, "__name__"):
                node_name = port.node_class.__name__
                port_name = getattr(port, "name", "unknown")
                return f"{node_name}.{port_name}"
            else:
                return str(port)
        except Exception:
            return f"<Port:{getattr(port, 'name', 'unknown')}>"

    def _get_node_name(self, node: Type["BaseNode"]) -> str:
        """Get a readable name for a node."""
        try:
            return getattr(node, "__name__", str(node))
        except Exception:
            return "<Node:unknown>"



"""Convenience utilities for reading Vellum files."""

from typing import TYPE_CHECKING, Optional

from vellum.utils.files.stream import stream_vellum_file
from vellum.utils.files.types import VellumFileTypes

if TYPE_CHECKING:
    from vellum.client import Vellum as VellumClient


def read_vellum_file(vellum_file: VellumFileTypes, *, vellum_client: Optional["VellumClient"] = None) -> bytes:
    """
    Convenience function that reads the entire file content into memory.

    This function is implemented using stream_vellum_file() under the hood.
    For large files (e.g., videos, large datasets), consider using
    stream_vellum_file() directly to avoid memory issues.

    Args:
        vellum_file: A VellumDocument, VellumImage, VellumAudio, or VellumVideo instance
        vellum_client: An optional Vellum client instance. If not provided, a default client will be created.

    Returns:
        bytes: The complete file content

    Example:
        ```python
        from vellum import VellumDocument
        from vellum.utils.files import read_vellum_file

        document = VellumDocument(src="data:text/plain;base64,SGVsbG8gV29ybGQ=")
        content = read_vellum_file(document)
        print(content.decode('utf-8'))  # "Hello World"
        ```
    """
    chunks = []
    with stream_vellum_file(vellum_file, vellum_client=vellum_client) as chunk_iter:
        for chunk in chunk_iter:
            chunks.append(chunk)
    return b"".join(chunks)



import logging
from uuid import UUID
from typing import TYPE_CHECKING, Any, Dict, Generic, Iterable, Literal, Optional, Type, Union, cast
from typing_extensions import TypeGuard

from pydantic import SerializationInfo, field_serializer, field_validator

from vellum.client.core.pydantic_utilities import UniversalBaseModel
from vellum.workflows.errors import WorkflowError
from vellum.workflows.outputs.base import BaseOutput
from vellum.workflows.references import ExternalInputReference
from vellum.workflows.types.definition import CodeResourceDefinition, serialize_type_encoder_with_id
from vellum.workflows.types.generics import InputsType, OutputsType, StateType

from ..triggers import BaseTrigger
from .node import (
    NodeExecutionFulfilledEvent,
    NodeExecutionInitiatedEvent,
    NodeExecutionPausedEvent,
    NodeExecutionRejectedEvent,
    NodeExecutionResumedEvent,
    NodeExecutionStreamingEvent,
)
from .stream import WorkflowEventGenerator
from .types import BaseEvent, default_serializer

if TYPE_CHECKING:
    from vellum.workflows.nodes.bases.base import BaseNode
    from vellum.workflows.workflows.base import BaseWorkflow

logger = logging.getLogger(__name__)


def _serialize_body_with_enricher(
    event: "_BaseWorkflowEvent", body: "_BaseWorkflowExecutionBody", info: SerializationInfo
) -> "_BaseWorkflowExecutionBody":
    """
    Helper function to serialize event body with optional event enrichment.

    This function is used by both WorkflowExecutionInitiatedEvent and WorkflowExecutionFulfilledEvent
    to apply event enrichment if an event_enricher is provided in the serialization context.
    """
    context = info.context if info and hasattr(info, "context") else {}
    if context and "event_enricher" in context and callable(context["event_enricher"]):
        try:
            enriched_event = context["event_enricher"](event)
            return enriched_event.body
        except Exception as e:
            logger.exception(f"Error in event_enricher: {e}")
            return body
    else:
        return body


class _BaseWorkflowExecutionBody(UniversalBaseModel):
    workflow_definition: Type["BaseWorkflow"]

    @field_serializer("workflow_definition")
    def serialize_workflow_definition(self, workflow_definition: Type, _info: Any) -> Dict[str, Any]:
        return serialize_type_encoder_with_id(workflow_definition)


class _BaseWorkflowEvent(BaseEvent):
    body: _BaseWorkflowExecutionBody

    @property
    def workflow_definition(self) -> Type["BaseWorkflow"]:
        return self.body.workflow_definition

    @property
    def monitoring_url(self) -> Optional[str]:
        """
        Get the monitoring URL for this workflow execution.

        Returns:
            The URL to view execution details in Vellum UI, or None if monitoring is disabled
            or context is not available.
        """
        return None


class NodeEventDisplayContext(UniversalBaseModel):
    input_display: Dict[str, UUID]
    output_display: Dict[str, UUID]
    port_display: Dict[str, UUID]
    subworkflow_display: Optional["WorkflowEventDisplayContext"] = None


class WorkflowEventDisplayContext(UniversalBaseModel):
    node_displays: Dict[UUID, NodeEventDisplayContext]
    workflow_inputs: Dict[str, UUID]
    workflow_outputs: Dict[str, UUID]


class WorkflowExecutionInitiatedBody(_BaseWorkflowExecutionBody, Generic[InputsType, StateType]):
    inputs: InputsType
    initial_state: Optional[StateType] = None

    # It is still the responsibility of the workflow server to populate this context. The SDK's
    # Workflow Runner will always leave this field None.
    #
    # It's conceivable in a future where all `id`s are agnostic to display and reside in a third location,
    # that the Workflow Runner can begin populating this field then.
    display_context: Optional[WorkflowEventDisplayContext] = None

    # This field will be populated during serialization by the serialize_body method
    workflow_version_exec_config: Optional[Any] = None

    # This field can be populated with arbitrary server metadata during event enrichment
    server_metadata: Optional[Dict[str, Any]] = None

    trigger: Optional[Type[BaseTrigger]] = None

    @field_serializer("inputs")
    def serialize_inputs(self, inputs: InputsType, _info: Any) -> Dict[str, Any]:
        return default_serializer(inputs)

    @field_serializer("initial_state")
    def serialize_initial_state(self, initial_state: Optional[StateType], _info: Any) -> Optional[Dict[str, Any]]:
        return default_serializer(initial_state)

    @field_serializer("trigger")
    def serialize_trigger(self, trigger: Optional[Type[BaseTrigger]], _info: Any) -> Optional[Dict[str, Any]]:
        if trigger is None:
            return None
        return serialize_type_encoder_with_id(trigger)


class WorkflowExecutionInitiatedEvent(_BaseWorkflowEvent, Generic[InputsType, StateType]):
    name: Literal["workflow.execution.initiated"] = "workflow.execution.initiated"
    body: WorkflowExecutionInitiatedBody[InputsType, StateType]

    @property
    def inputs(self) -> InputsType:
        return self.body.inputs

    @property
    def initial_state(self) -> Optional[StateType]:
        return self.body.initial_state

    @field_serializer("body")
    def serialize_body(
        self, body: WorkflowExecutionInitiatedBody[InputsType, StateType], info: SerializationInfo
    ) -> WorkflowExecutionInitiatedBody[InputsType, StateType]:
        return cast(
            WorkflowExecutionInitiatedBody[InputsType, StateType], _serialize_body_with_enricher(self, body, info)
        )


class WorkflowExecutionStreamingBody(_BaseWorkflowExecutionBody):
    output: BaseOutput

    @field_serializer("output")
    def serialize_output(self, output: BaseOutput, _info: Any) -> Dict[str, Any]:
        return default_serializer(output)


class WorkflowExecutionStreamingEvent(_BaseWorkflowEvent):
    name: Literal["workflow.execution.streaming"] = "workflow.execution.streaming"
    body: WorkflowExecutionStreamingBody

    @property
    def output(self) -> BaseOutput:
        return self.body.output


class WorkflowExecutionFulfilledBody(_BaseWorkflowExecutionBody, Generic[OutputsType, StateType]):
    outputs: OutputsType
    final_state: Optional[StateType] = None

    # This field can be populated with arbitrary server metadata during event enrichment
    server_metadata: Optional[Dict[str, Any]] = None

    @field_serializer("outputs")
    def serialize_outputs(self, outputs: OutputsType, _info: Any) -> Dict[str, Any]:
        return default_serializer(outputs)

    @field_serializer("final_state")
    def serialize_final_state(self, final_state: Optional[StateType], _info: Any) -> Optional[Dict[str, Any]]:
        if final_state is None:
            return None
        state_dict = {k: v for k, v in final_state if k != "meta"}
        return default_serializer(state_dict)


class WorkflowExecutionFulfilledEvent(_BaseWorkflowEvent, Generic[OutputsType, StateType]):
    name: Literal["workflow.execution.fulfilled"] = "workflow.execution.fulfilled"
    body: WorkflowExecutionFulfilledBody[OutputsType, StateType]

    @property
    def outputs(self) -> OutputsType:
        return self.body.outputs

    @property
    def final_state(self) -> Optional[StateType]:
        return self.body.final_state

    @field_serializer("body")
    def serialize_body(
        self, body: WorkflowExecutionFulfilledBody[OutputsType, StateType], info: SerializationInfo
    ) -> WorkflowExecutionFulfilledBody[OutputsType, StateType]:
        return cast(
            WorkflowExecutionFulfilledBody[OutputsType, StateType], _serialize_body_with_enricher(self, body, info)
        )


class WorkflowExecutionRejectedBody(_BaseWorkflowExecutionBody):
    error: WorkflowError
    stacktrace: Optional[str] = None


class WorkflowExecutionRejectedEvent(_BaseWorkflowEvent):
    name: Literal["workflow.execution.rejected"] = "workflow.execution.rejected"
    body: WorkflowExecutionRejectedBody

    @property
    def error(self) -> WorkflowError:
        return self.body.error

    @property
    def stacktrace(self) -> Optional[str]:
        return self.body.stacktrace


class WorkflowExecutionPausedBody(_BaseWorkflowExecutionBody):
    external_inputs: Iterable[ExternalInputReference]


class WorkflowExecutionPausedEvent(_BaseWorkflowEvent):
    name: Literal["workflow.execution.paused"] = "workflow.execution.paused"
    body: WorkflowExecutionPausedBody

    @property
    def external_inputs(self) -> Iterable[ExternalInputReference]:
        return self.body.external_inputs


class WorkflowExecutionResumedBody(_BaseWorkflowExecutionBody):
    pass


class WorkflowExecutionResumedEvent(_BaseWorkflowEvent):
    name: Literal["workflow.execution.resumed"] = "workflow.execution.resumed"
    body: WorkflowExecutionResumedBody


class WorkflowExecutionSnapshottedBody(_BaseWorkflowExecutionBody, Generic[StateType]):
    state: StateType
    edited_by: Optional[Type["BaseNode"]] = None

    @field_validator("edited_by", mode="before")
    @classmethod
    def validate_edited_by(cls, value: Any) -> Any:
        if value is None:
            return None
        value = CodeResourceDefinition.model_validate(value)
        return value.decode()

    @field_serializer("state")
    def serialize_state(self, state: StateType, _info: Any) -> Dict[str, Any]:
        return default_serializer(state)

    @field_serializer("edited_by")
    def serialize_edited_by(self, edited_by: Optional[Type["BaseNode"]], _info: Any) -> Optional[Dict[str, Any]]:
        if edited_by is None:
            return None
        return serialize_type_encoder_with_id(edited_by)


class WorkflowExecutionSnapshottedEvent(_BaseWorkflowEvent, Generic[StateType]):
    name: Literal["workflow.execution.snapshotted"] = "workflow.execution.snapshotted"
    body: WorkflowExecutionSnapshottedBody[StateType]

    @property
    def state(self) -> StateType:
        return self.body.state


GenericWorkflowEvent = Union[
    WorkflowExecutionStreamingEvent,
    WorkflowExecutionRejectedEvent,
    WorkflowExecutionPausedEvent,
    WorkflowExecutionResumedEvent,
    NodeExecutionInitiatedEvent,
    NodeExecutionStreamingEvent,
    NodeExecutionFulfilledEvent,
    NodeExecutionRejectedEvent,
    NodeExecutionPausedEvent,
    NodeExecutionResumedEvent,
]

WorkflowEvent = Union[
    GenericWorkflowEvent,
    WorkflowExecutionInitiatedEvent,
    WorkflowExecutionFulfilledEvent,
    WorkflowExecutionSnapshottedEvent,
]

WorkflowEventStream = WorkflowEventGenerator[WorkflowEvent]

WorkflowExecutionEvent = Union[
    WorkflowExecutionInitiatedEvent,
    WorkflowExecutionStreamingEvent,
    WorkflowExecutionRejectedEvent,
    WorkflowExecutionPausedEvent,
    WorkflowExecutionResumedEvent,
    WorkflowExecutionFulfilledEvent,
    WorkflowExecutionSnapshottedEvent,
]

TerminalWorkflowExecutionEvent = Union[
    WorkflowExecutionFulfilledEvent,
    WorkflowExecutionRejectedEvent,
    WorkflowExecutionPausedEvent,
]


def is_workflow_event(event: WorkflowEvent) -> TypeGuard[WorkflowExecutionEvent]:
    return (
        event.name == "workflow.execution.initiated"
        or event.name == "workflow.execution.fulfilled"
        or event.name == "workflow.execution.streaming"
        or event.name == "workflow.execution.snapshotted"
        or event.name == "workflow.execution.paused"
        or event.name == "workflow.execution.resumed"
        or event.name == "workflow.execution.rejected"
    )


def is_terminal_workflow_execution_event(event: WorkflowEvent) -> TypeGuard[TerminalWorkflowExecutionEvent]:
    return (
        event.name == "workflow.execution.fulfilled"
        or event.name == "workflow.execution.rejected"
        or event.name == "workflow.execution.paused"
    )



"""Tests for IntegrationTrigger."""

import pytest

from vellum.workflows.constants import VellumIntegrationProviderType
from vellum.workflows.references.trigger import TriggerAttributeReference
from vellum.workflows.triggers.integration import IntegrationTrigger


def test_requires_config():
    """Trigger classes must declare Config with required fields."""

    # This should fail - no Config class
    with pytest.raises(TypeError, match="Config"):

        class BadTrigger1(IntegrationTrigger):
            message: str
            # Missing Config!

    # This should fail - incomplete Config
    with pytest.raises(TypeError, match="provider"):

        class BadTrigger2(IntegrationTrigger):
            message: str

            class Config(IntegrationTrigger.Config):
                integration_name = "SLACK"
                slug = "slack_new_message"
                # Missing provider!

    # This should work
    class GoodTrigger(IntegrationTrigger):
        message: str
        user: str

        class Config(IntegrationTrigger.Config):
            provider = VellumIntegrationProviderType.COMPOSIO
            integration_name = "SLACK"
            slug = "slack_new_message"

    assert hasattr(GoodTrigger, "Config")


def test_top_level_annotations_create_references():
    """Top-level type annotations (webhook event attributes) automatically create TriggerAttributeReference."""

    class SlackTrigger(IntegrationTrigger):
        message: str
        user: str
        timestamp: float

        class Config(IntegrationTrigger.Config):
            provider = VellumIntegrationProviderType.COMPOSIO
            integration_name = "SLACK"
            slug = "slack_new_message"

    # Should auto-create references from annotations
    assert isinstance(SlackTrigger.message, TriggerAttributeReference)
    assert isinstance(SlackTrigger.user, TriggerAttributeReference)
    assert isinstance(SlackTrigger.timestamp, TriggerAttributeReference)
    assert SlackTrigger.message.types == (str,)
    assert SlackTrigger.timestamp.types == (float,)

    # Undeclared attributes should raise AttributeError
    with pytest.raises(AttributeError):
        _ = SlackTrigger.undefined_attribute


def test_attribute_ids_include_class_name():
    """Attribute IDs should include class name (like nodes)."""

    class Trigger1(IntegrationTrigger):
        message: str

        class Config(IntegrationTrigger.Config):
            provider = VellumIntegrationProviderType.COMPOSIO
            integration_name = "SLACK"
            slug = "slack_new_message"

    class Trigger2(IntegrationTrigger):
        message: str

        class Config(IntegrationTrigger.Config):
            provider = VellumIntegrationProviderType.COMPOSIO
            integration_name = "SLACK"
            slug = "slack_new_message"

    # Different class names = different IDs (like nodes)
    # Type ignore: mypy sees message as str, but it's actually TriggerAttributeReference at class level
    assert Trigger1.message.id != Trigger2.message.id  # type: ignore[attr-defined]


def test_populates_dynamic_attributes():
    """Trigger dynamically populates attributes from event_data keys."""

    class GithubPushTrigger(IntegrationTrigger):
        repository: str
        branch: str
        commits: list

        class Config(IntegrationTrigger.Config):
            provider = VellumIntegrationProviderType.COMPOSIO
            integration_name = "GITHUB"
            slug = "github_push_event"

    event_data = {
        "repository": "vellum-ai/workflows",
        "branch": "main",
        "commits": ["abc123", "def456"],
    }

    trigger = GithubPushTrigger(**event_data)

    assert trigger.repository == "vellum-ai/workflows"
    assert trigger.branch == "main"
    assert trigger.commits == ["abc123", "def456"]


def test_to_trigger_attribute_values():
    """to_trigger_attribute_values returns correct attribute mappings."""

    class SlackTrigger(IntegrationTrigger):
        message: str
        channel: str

        class Config(IntegrationTrigger.Config):
            provider = VellumIntegrationProviderType.COMPOSIO
            integration_name = "SLACK"
            slug = "slack_new_message"

    event_data = {"message": "Hello", "channel": "C123"}
    trigger = SlackTrigger(**event_data)

    attr_values = trigger.to_trigger_attribute_values()

    assert len(attr_values) == 2
    for key in attr_values.keys():
        assert isinstance(key, TriggerAttributeReference)
    assert set(attr_values.values()) == {"Hello", "C123"}


def test_empty_event_data():
    """Trigger handles empty event data gracefully."""

    class SlackTrigger(IntegrationTrigger):
        message: str

        class Config(IntegrationTrigger.Config):
            provider = VellumIntegrationProviderType.COMPOSIO
            integration_name = "SLACK"
            slug = "slack_new_message"

    trigger = SlackTrigger()

    attr_values = trigger.to_trigger_attribute_values()
    assert attr_values == {}



from .api_node import BaseAPINodeDisplay
from .base_adornment_node import BaseAdornmentNodeDisplay
from .code_execution_node import BaseCodeExecutionNodeDisplay
from .conditional_node import BaseConditionalNodeDisplay
from .error_node import BaseErrorNodeDisplay
from .final_output_node import BaseFinalOutputNodeDisplay
from .function_node import FunctionNodeDisplay
from .guardrail_node import BaseGuardrailNodeDisplay
from .inline_prompt_node import BaseInlinePromptNodeDisplay
from .inline_subworkflow_node import BaseInlineSubworkflowNodeDisplay
from .map_node import BaseMapNodeDisplay
from .merge_node import BaseMergeNodeDisplay
from .note_node import BaseNoteNodeDisplay
from .prompt_deployment_node import BasePromptDeploymentNodeDisplay
from .retry_node import BaseRetryNodeDisplay
from .search_node import BaseSearchNodeDisplay
from .subworkflow_deployment_node import BaseSubworkflowDeploymentNodeDisplay
from .templating_node import BaseTemplatingNodeDisplay
from .try_node import BaseTryNodeDisplay

# All node display classes must be imported here to be registered in BaseNodeDisplay's node display registry
__all__ = [
    "BaseAdornmentNodeDisplay",
    "BaseAPINodeDisplay",
    "BaseCodeExecutionNodeDisplay",
    "BaseConditionalNodeDisplay",
    "BaseErrorNodeDisplay",
    "BaseFinalOutputNodeDisplay",
    "FunctionNodeDisplay",
    "BaseGuardrailNodeDisplay",
    "BaseInlinePromptNodeDisplay",
    "BaseInlineSubworkflowNodeDisplay",
    "BaseMapNodeDisplay",
    "BaseMergeNodeDisplay",
    "BaseNoteNodeDisplay",
    "BasePromptDeploymentNodeDisplay",
    "BaseRetryNodeDisplay",
    "BaseSearchNodeDisplay",
    "BaseSubworkflowDeploymentNodeDisplay",
    "BaseTemplatingNodeDisplay",
    "BaseTryNodeDisplay",
]



import pytest
from uuid import UUID

from vellum.workflows.nodes.bases import BaseNode
from vellum.workflows.ports.port import Port
from vellum.workflows.references.constant import ConstantValueReference
from vellum_ee.workflows.display.editor.types import NodeDisplayData, NodeDisplayPosition
from vellum_ee.workflows.display.nodes.base_node_display import BaseNodeDisplay
from vellum_ee.workflows.display.nodes.get_node_display_class import get_node_display_class
from vellum_ee.workflows.display.types import WorkflowDisplayContext


@pytest.fixture
def node_with_implicit_properties():
    class MyNode(BaseNode):
        pass

    class MyNodeDisplay(BaseNodeDisplay[MyNode]):
        pass

    expected_id = UUID("8628fe5e-620d-4935-92f1-a9ed7ed7fd92")

    return MyNodeDisplay, expected_id


@pytest.fixture
def node_with_explicit_properties():
    explicit_id = UUID("a422f67a-1d37-43f0-bdfc-1e4618c9496d")

    class MyNode(BaseNode):
        pass

    class MyNodeDisplay(BaseNodeDisplay[MyNode]):
        node_id = explicit_id

    return MyNodeDisplay, explicit_id


@pytest.fixture(
    params=[
        "node_with_implicit_properties",
        "node_with_explicit_properties",
    ]
)
def node_info(request):
    return request.getfixturevalue(request.param)


def test_get_id(node_info):
    node_display, expected_id = node_info

    assert node_display().node_id == expected_id
    assert node_display.infer_node_class().__id__ == expected_id


def test_serialize_condition__accessor_expression():
    # GIVEN a node with an accessor expression in a Port
    class MyNode(BaseNode):

        class Ports(BaseNode.Ports):
            foo = Port.on_if(ConstantValueReference({"hello": "world"})["hello"])

    # WHEN we serialize the node
    node_display_class = get_node_display_class(MyNode)
    data = node_display_class().serialize(WorkflowDisplayContext())

    # THEN the condition should be serialized correctly
    assert data["ports"] == [
        {
            "id": "c5a19e2d-3476-4d51-ae8f-a00f101dec30",
            "name": "foo",
            "type": "IF",
            "expression": {
                "type": "BINARY_EXPRESSION",
                "lhs": {
                    "type": "CONSTANT_VALUE",
                    "value": {
                        "type": "JSON",
                        "value": {"hello": "world"},
                    },
                },
                "operator": "accessField",
                "rhs": {"type": "CONSTANT_VALUE", "value": {"type": "STRING", "value": "hello"}},
            },
        }
    ]


def test_serialize_display_data():
    # GIVEN a node with an accessor expression in a Port
    class MyNode(BaseNode):
        """I hope this works"""

        pass

    # WHEN we serialize the node
    node_display_class = get_node_display_class(MyNode)
    data = node_display_class().serialize(WorkflowDisplayContext())

    # THEN the condition should be serialized correctly
    assert data["display_data"] == {
        "position": {"x": 0.0, "y": 0.0},
        "comment": {"expanded": True, "value": "I hope this works"},
    }


def test_serialize_node_label_with_pascal_case():
    """
    Tests that a node with PascalCase name serializes with proper title case label.
    """

    # GIVEN a node with a PascalCase name that includes common patterns
    class MyCustomNode(BaseNode):
        pass

    # WHEN we serialize the node
    node_display_class = get_node_display_class(MyCustomNode)
    data = node_display_class().serialize(WorkflowDisplayContext())

    # THEN the label should be converted to proper title case with spaces
    assert data["label"] == "My Custom Node"


def test_serialize_display_data_with_icon_and_color():
    """
    Tests that nodes with icon and color serialize display_data correctly.
    """

    # GIVEN a node with icon and color in display data
    class MyNode(BaseNode):
        pass

    class MyNodeDisplay(BaseNodeDisplay[MyNode]):
        display_data = NodeDisplayData(
            position=NodeDisplayPosition(x=100, y=200), icon="vellum:icon:star", color="navy"
        )

    # WHEN we serialize the node
    data = MyNodeDisplay().serialize(WorkflowDisplayContext())

    # THEN the display_data should include icon and color
    assert data["display_data"] == {"position": {"x": 100, "y": 200}, "icon": "vellum:icon:star", "color": "navy"}


def test_serialize_display_data_with_various_icon_formats():
    """
    Tests that different icon formats are serialized correctly.
    """

    # GIVEN a node with a vellum icon format
    class MyNode(BaseNode):
        pass

    class MyNodeDisplay(BaseNodeDisplay[MyNode]):
        display_data = NodeDisplayData(icon="vellum:icon:home")

    # WHEN we serialize the node
    data = MyNodeDisplay().serialize(WorkflowDisplayContext())

    # THEN the icon should be preserved as-is
    display_data = data["display_data"]
    assert isinstance(display_data, dict)
    assert display_data["icon"] == "vellum:icon:home"


def test_serialize_display_data_with_navy_color():
    """
    Tests that navy color values are serialized correctly.
    """

    # GIVEN a node with a navy color
    class MyNode(BaseNode):
        pass

    class MyNodeDisplay(BaseNodeDisplay[MyNode]):
        display_data = NodeDisplayData(color="navy")

    # WHEN we serialize the node
    data = MyNodeDisplay().serialize(WorkflowDisplayContext())

    # THEN the color should be preserved as-is
    display_data = data["display_data"]
    assert isinstance(display_data, dict)
    assert display_data["color"] == "navy"


def test_serialize_basenode_with_display_class():
    """Tests that BaseNode.Display class attributes serialize correctly."""

    class MyNode(BaseNode):
        class Display:
            icon = "vellum:icon:gear"
            color = "purple"

    node_display_class = get_node_display_class(MyNode)
    data = node_display_class().serialize(WorkflowDisplayContext())

    display_data = data["display_data"]
    assert isinstance(display_data, dict)
    assert display_data["icon"] == "vellum:icon:gear"
    assert display_data["color"] == "purple"


def test_serialize_explicit_display_data_overrides_display_class():
    """Tests that BaseNodeDisplay's explicit display_data takes precedence over Display class."""

    class MyNode(BaseNode):
        class Display:
            icon = "vellum:icon:check"
            color = "gold"

    class MyNodeDisplay(BaseNodeDisplay[MyNode]):
        display_data = NodeDisplayData(
            position=NodeDisplayPosition(x=100, y=200),
            icon="vellum:icon:times",
            color="navy",
        )

    data = MyNodeDisplay().serialize(WorkflowDisplayContext())

    display_data = data["display_data"]
    assert isinstance(display_data, dict)
    assert display_data["icon"] == "vellum:icon:times"  # BaseNodeDisplay overrides
    assert display_data["color"] == "navy"  # BaseNodeDisplay overrides
    assert display_data["position"] == {"x": 100, "y": 200}


def test_serialize_display_class_used_as_fallback():
    """Tests that Display class attributes are used when BaseNodeDisplay doesn't specify them."""

    class MyNode(BaseNode):
        class Display:
            icon = "vellum:icon:star"
            color = "green"

    class MyNodeDisplay(BaseNodeDisplay[MyNode]):
        display_data = NodeDisplayData(
            position=NodeDisplayPosition(x=50, y=75),
            # Not specifying icon or color - should fall back to Display class
        )

    data = MyNodeDisplay().serialize(WorkflowDisplayContext())

    display_data = data["display_data"]
    assert isinstance(display_data, dict)
    assert display_data["icon"] == "vellum:icon:star"  # Falls back to Display class
    assert display_data["color"] == "green"  # Falls back to Display class
    assert display_data["position"] == {"x": 50, "y": 75}  # From BaseNodeDisplay



from uuid import UUID
from typing import Any, ClassVar, Generic, List, Optional, TypeVar

from vellum.workflows.nodes.displayable import MergeNode
from vellum.workflows.types.core import JsonObject
from vellum.workflows.utils.uuids import uuid4_from_hash
from vellum_ee.workflows.display.nodes.base_node_display import BaseNodeDisplay
from vellum_ee.workflows.display.types import WorkflowDisplayContext

_MergeNodeType = TypeVar("_MergeNodeType", bound=MergeNode)


class BaseMergeNodeDisplay(BaseNodeDisplay[_MergeNodeType], Generic[_MergeNodeType]):
    target_handle_ids: ClassVar[List[UUID]]

    def __init__(self):
        super().__init__()
        self._target_handle_iterator = 0

    def serialize(self, display_context: WorkflowDisplayContext, **_kwargs: Any) -> JsonObject:
        node = self._node
        node_id = self.node_id

        merged_source_ports = [
            source_node_port
            for (source_node_port, target_node), _ in display_context.edge_displays.items()
            if target_node.__id__ == self.node_id
        ]

        target_handle_ids = self.get_target_handle_ids()

        if target_handle_ids is None:
            target_handle_ids = [
                uuid4_from_hash(f"{node_id}|target_handle|{source_node_port.node_class.__id__}")
                for source_node_port in merged_source_ports
            ]
        elif len(target_handle_ids) != len(merged_source_ports):
            raise ValueError("If you explicitly specify target_handle_ids, you must specify one for each incoming edge")

        return {
            "id": str(node_id),
            "type": "MERGE",
            "inputs": [],
            "data": {
                "label": self.label,
                "merge_strategy": node.Trigger.merge_behavior.value,
                "target_handles": [{"id": str(target_handle_id)} for target_handle_id in target_handle_ids],
                "source_handle_id": str(self.get_source_handle_id(display_context.port_displays)),
            },
            **self.serialize_generic_fields(display_context, exclude=["outputs"]),
        }

    def get_target_handle_ids(self) -> Optional[List[UUID]]:
        return self._get_explicit_node_display_attr("target_handle_ids", List[UUID])

    def get_target_handle_id_by_source_node_id(self, source_node_id: UUID) -> UUID:
        target_handle_ids = self.get_target_handle_ids()
        if target_handle_ids is None:
            return uuid4_from_hash(f"{self.node_id}|target_handle|{source_node_id}")

        # Edges call this method to know which handle to connect to
        # We use the order of the edges to determine the handle id. This is quite brittle to the order of the
        # edges matching the order of the Merge Node's data.target_handles attribute. We should look into a
        # longer term solution, or cutover Merge Nodes to Generic Nodes soon
        target_handle_id = target_handle_ids[self._target_handle_iterator]
        self._target_handle_iterator += 1
        if self._target_handle_iterator >= len(self.target_handle_ids):
            self._target_handle_iterator = 0
        return target_handle_id



# This file was auto-generated by Fern from our API Definition.

import typing
from json.decoder import JSONDecodeError

from ... import core
from ...core.api_error import ApiError
from ...core.client_wrapper import AsyncClientWrapper, SyncClientWrapper
from ...core.http_response import AsyncHttpResponse, HttpResponse
from ...core.jsonable_encoder import jsonable_encoder
from ...core.pydantic_utilities import parse_obj_as
from ...core.request_options import RequestOptions
from ...types.uploaded_file_read import UploadedFileRead

# this is used as the default value for optional parameters
OMIT = typing.cast(typing.Any, ...)


class RawUploadedFilesClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._client_wrapper = client_wrapper

    def create(
        self, *, file: core.File, request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[UploadedFileRead]:
        """
        Upload a file to be used in the Workspace

        Parameters
        ----------
        file : core.File
            See core.File for more documentation

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[UploadedFileRead]

        """
        _response = self._client_wrapper.httpx_client.request(
            "v1/uploaded-files",
            base_url=self._client_wrapper.get_environment().default,
            method="POST",
            data={},
            files={
                "file": file,
            },
            request_options=request_options,
            omit=OMIT,
            force_multipart=True,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    UploadedFileRead,
                    parse_obj_as(
                        type_=UploadedFileRead,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def retrieve(
        self,
        id: str,
        *,
        expiry_seconds: typing.Optional[int] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[UploadedFileRead]:
        """
        Retrieve a previously uploaded file by its ID

        Parameters
        ----------
        id : str
            A UUID string identifying this uploaded file.

        expiry_seconds : typing.Optional[int]
            The number of seconds until the signed URL expires. Must be > 0 and <= 2592000 (30 days). Non-numeric or out-of-range values fall back to the default (604800 seconds or 7 days).

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[UploadedFileRead]

        """
        _response = self._client_wrapper.httpx_client.request(
            f"v1/uploaded-files/{jsonable_encoder(id)}",
            base_url=self._client_wrapper.get_environment().default,
            method="GET",
            params={
                "expiry_seconds": expiry_seconds,
            },
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    UploadedFileRead,
                    parse_obj_as(
                        type_=UploadedFileRead,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def update(
        self, id: str, *, file: core.File, request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[UploadedFileRead]:
        """
        Update an uploaded file by its ID

        Parameters
        ----------
        id : str
            A UUID string identifying this uploaded file.

        file : core.File
            See core.File for more documentation

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[UploadedFileRead]

        """
        _response = self._client_wrapper.httpx_client.request(
            f"v1/uploaded-files/{jsonable_encoder(id)}",
            base_url=self._client_wrapper.get_environment().default,
            method="PUT",
            data={},
            files={
                "file": file,
            },
            request_options=request_options,
            omit=OMIT,
            force_multipart=True,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    UploadedFileRead,
                    parse_obj_as(
                        type_=UploadedFileRead,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)


class AsyncRawUploadedFilesClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._client_wrapper = client_wrapper

    async def create(
        self, *, file: core.File, request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[UploadedFileRead]:
        """
        Upload a file to be used in the Workspace

        Parameters
        ----------
        file : core.File
            See core.File for more documentation

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[UploadedFileRead]

        """
        _response = await self._client_wrapper.httpx_client.request(
            "v1/uploaded-files",
            base_url=self._client_wrapper.get_environment().default,
            method="POST",
            data={},
            files={
                "file": file,
            },
            request_options=request_options,
            omit=OMIT,
            force_multipart=True,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    UploadedFileRead,
                    parse_obj_as(
                        type_=UploadedFileRead,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def retrieve(
        self,
        id: str,
        *,
        expiry_seconds: typing.Optional[int] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[UploadedFileRead]:
        """
        Retrieve a previously uploaded file by its ID

        Parameters
        ----------
        id : str
            A UUID string identifying this uploaded file.

        expiry_seconds : typing.Optional[int]
            The number of seconds until the signed URL expires. Must be > 0 and <= 2592000 (30 days). Non-numeric or out-of-range values fall back to the default (604800 seconds or 7 days).

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[UploadedFileRead]

        """
        _response = await self._client_wrapper.httpx_client.request(
            f"v1/uploaded-files/{jsonable_encoder(id)}",
            base_url=self._client_wrapper.get_environment().default,
            method="GET",
            params={
                "expiry_seconds": expiry_seconds,
            },
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    UploadedFileRead,
                    parse_obj_as(
                        type_=UploadedFileRead,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def update(
        self, id: str, *, file: core.File, request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[UploadedFileRead]:
        """
        Update an uploaded file by its ID

        Parameters
        ----------
        id : str
            A UUID string identifying this uploaded file.

        file : core.File
            See core.File for more documentation

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[UploadedFileRead]

        """
        _response = await self._client_wrapper.httpx_client.request(
            f"v1/uploaded-files/{jsonable_encoder(id)}",
            base_url=self._client_wrapper.get_environment().default,
            method="PUT",
            data={},
            files={
                "file": file,
            },
            request_options=request_options,
            omit=OMIT,
            force_multipart=True,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    UploadedFileRead,
                    parse_obj_as(
                        type_=UploadedFileRead,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)



# This file was auto-generated by Fern from our API Definition.

from __future__ import annotations

import datetime as dt
import typing

import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel, update_forward_refs
from .api_version_enum import ApiVersionEnum
from .node_execution_rejected_body import NodeExecutionRejectedBody


class NodeExecutionRejectedEvent(UniversalBaseModel):
    parent: typing.Optional["ParentContext"] = None
    links: typing.Optional[typing.List["SpanLink"]] = None
    name: typing.Literal["node.execution.rejected"] = "node.execution.rejected"
    body: NodeExecutionRejectedBody
    id: str
    timestamp: dt.datetime
    api_version: typing.Optional[ApiVersionEnum] = None
    trace_id: str
    span_id: str

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow


from .api_request_parent_context import ApiRequestParentContext  # noqa: E402, F401, I001
from .external_parent_context import ExternalParentContext  # noqa: E402, F401, I001
from .integration_trigger_context import IntegrationTriggerContext  # noqa: E402, F401, I001
from .node_parent_context import NodeParentContext  # noqa: E402, F401, I001
from .prompt_deployment_parent_context import PromptDeploymentParentContext  # noqa: E402, F401, I001
from .scheduled_trigger_context import ScheduledTriggerContext  # noqa: E402, F401, I001
from .span_link import SpanLink  # noqa: E402, F401, I001
from .workflow_deployment_parent_context import WorkflowDeploymentParentContext  # noqa: E402, F401, I001
from .workflow_parent_context import WorkflowParentContext  # noqa: E402, F401, I001
from .workflow_sandbox_parent_context import WorkflowSandboxParentContext  # noqa: E402, F401, I001
from .parent_context import ParentContext  # noqa: E402, F401, I001

update_forward_refs(NodeExecutionRejectedEvent)



import json
import logging
from typing import Any, Callable, Dict, Iterator, List, Optional, Type, Union, cast

from pydash import snake_case

from vellum import ChatMessage, PromptBlock
from vellum.client.types.array_chat_message_content import ArrayChatMessageContent
from vellum.client.types.array_chat_message_content_item import ArrayChatMessageContentItem
from vellum.client.types.function_call_chat_message_content import FunctionCallChatMessageContent
from vellum.client.types.function_call_chat_message_content_value import FunctionCallChatMessageContentValue
from vellum.client.types.prompt_output import PromptOutput
from vellum.client.types.prompt_parameters import PromptParameters
from vellum.client.types.prompt_settings import PromptSettings
from vellum.client.types.string_chat_message_content import StringChatMessageContent
from vellum.client.types.variable_prompt_block import VariablePromptBlock
from vellum.utils.json_encoder import VellumJsonEncoder
from vellum.workflows.descriptors.base import BaseDescriptor
from vellum.workflows.errors.types import WorkflowErrorCode
from vellum.workflows.exceptions import NodeException
from vellum.workflows.expressions.concat import ConcatExpression
from vellum.workflows.inputs import BaseInputs
from vellum.workflows.integrations.composio_service import ComposioService
from vellum.workflows.integrations.mcp_service import MCPService
from vellum.workflows.integrations.vellum_integration_service import VellumIntegrationService
from vellum.workflows.nodes.bases import BaseNode
from vellum.workflows.nodes.core.inline_subworkflow_node.node import InlineSubworkflowNode
from vellum.workflows.nodes.displayable.inline_prompt_node.node import InlinePromptNode
from vellum.workflows.nodes.displayable.subworkflow_deployment_node.node import SubworkflowDeploymentNode
from vellum.workflows.nodes.displayable.tool_calling_node.state import ToolCallingState
from vellum.workflows.outputs.base import BaseOutput
from vellum.workflows.ports.port import Port
from vellum.workflows.state import BaseState
from vellum.workflows.types.core import EntityInputsInterface, MergeBehavior
from vellum.workflows.types.definition import (
    ComposioToolDefinition,
    DeploymentDefinition,
    MCPServer,
    MCPToolDefinition,
    Tool,
    ToolBase,
    VellumIntegrationToolDefinition,
)
from vellum.workflows.types.generics import is_workflow_class
from vellum.workflows.utils.functions import compile_mcp_tool_definition, get_mcp_tool_name

CHAT_HISTORY_VARIABLE = "chat_history"


logger = logging.getLogger(__name__)


class FunctionCallNodeMixin:
    """Mixin providing common functionality for nodes that handle function calls."""

    function_call_output: List[PromptOutput]

    def _handle_tool_exception(self, e: Exception, tool_type: str, tool_name: str) -> None:
        """
        Re-raise exceptions with contextual information while preserving NodeException details.

        Args:
            e: The caught exception
            tool_type: Type of tool (e.g., "function", "MCP tool", "Vellum Integration tool")
            tool_name: Name of the tool that failed
        """
        if isinstance(e, NodeException):
            # Preserve original error code and raw_data while adding context
            raise NodeException(
                message=f"Error executing {tool_type} '{tool_name}': {e.message}",
                code=e.code,
                raw_data=e.raw_data,
            ) from e
        else:
            raise NodeException(
                message=f"Error executing {tool_type} '{tool_name}': {str(e)}",
                code=WorkflowErrorCode.NODE_EXECUTION,
            ) from e

    def _extract_function_arguments(self) -> dict:
        """Extract arguments from function call output."""
        current_index = getattr(self, "state").current_prompt_output_index
        if self.function_call_output and len(self.function_call_output) > current_index:
            function_call = self.function_call_output[current_index]
            if function_call.type == "FUNCTION_CALL" and function_call.value is not None:
                return function_call.value.arguments or {}
        return {}

    def _extract_function_call_id(self) -> Optional[str]:
        """Extract function call ID from function call output."""
        current_index = getattr(self, "state").current_prompt_output_index
        if self.function_call_output and len(self.function_call_output) > current_index:
            function_call = self.function_call_output[current_index]
            if function_call.type == "FUNCTION_CALL" and function_call.value is not None:
                return function_call.value.id
        return None

    def _add_function_result_to_chat_history(self, result: Any, state: ToolCallingState) -> None:
        """Add function execution result to chat history."""
        function_call_id = self._extract_function_call_id()
        state.chat_history.append(
            ChatMessage(
                role="FUNCTION",
                content=StringChatMessageContent(value=json.dumps(result, cls=VellumJsonEncoder)),
                source=function_call_id,
            )
        )
        with state.__quiet__():
            state.current_function_calls_processed += 1
            state.current_prompt_output_index += 1


class ToolPromptNode(InlinePromptNode[ToolCallingState]):
    max_prompt_iterations: Optional[int] = 25

    class Trigger(InlinePromptNode.Trigger):
        merge_behavior = MergeBehavior.AWAIT_ATTRIBUTES

    def run(self) -> Iterator[BaseOutput]:
        if self.max_prompt_iterations is not None and self.state.prompt_iterations >= self.max_prompt_iterations:
            max_iterations_message = f"Maximum number of prompt iterations `{self.max_prompt_iterations}` reached."
            raise NodeException(message=max_iterations_message, code=WorkflowErrorCode.NODE_EXECUTION)

        generator = super().run()
        with self.state.__quiet__():
            self.state.current_prompt_output_index = 0
            self.state.current_function_calls_processed = 0
            self.state.prompt_iterations += 1
        for output in generator:
            if output.name == InlinePromptNode.Outputs.results.name and output.value:
                prompt_outputs = cast(List[PromptOutput], output.value)
                chat_contents: List[ArrayChatMessageContentItem] = []
                for prompt_output in prompt_outputs:
                    if prompt_output.type == "STRING":
                        chat_contents.append(StringChatMessageContent(value=prompt_output.value))
                    elif prompt_output.type == "FUNCTION_CALL" and prompt_output.value:
                        raw_function_call = prompt_output.value.model_dump()
                        if "state" in raw_function_call:
                            del raw_function_call["state"]
                        chat_contents.append(
                            FunctionCallChatMessageContent(
                                value=FunctionCallChatMessageContentValue.model_validate(raw_function_call)
                            )
                        )

                if len(chat_contents) == 1:
                    if chat_contents[0].type == "STRING":
                        self.state.chat_history.append(ChatMessage(role="ASSISTANT", text=chat_contents[0].value))
                    else:
                        self.state.chat_history.append(ChatMessage(role="ASSISTANT", content=chat_contents[0]))
                else:
                    self.state.chat_history.append(
                        ChatMessage(role="ASSISTANT", content=ArrayChatMessageContent(value=chat_contents))
                    )

            yield output


class RouterNode(BaseNode[ToolCallingState]):
    """Router node that handles routing to function nodes based on outputs."""

    __exclude_from_monitoring__: bool = True  # Exclude from monitoring views


class DynamicSubworkflowDeploymentNode(SubworkflowDeploymentNode[ToolCallingState], FunctionCallNodeMixin):
    """Node that executes a deployment definition with function call output."""

    def run(self) -> Iterator[BaseOutput]:
        arguments = self._extract_function_arguments()

        # Mypy doesn't like instance assignments of class attributes. It's safe in our case tho bc it's what
        # we do in the `__init__` method. Long term, instead of the function_call_output attribute above, we
        # want to do:
        # ```python
        # subworkflow_inputs = tool_prompt_node.Outputs.results[0]['value']['arguments'].if_(
        #     tool_prompt_node.Outputs.results[0]['type'].equals('FUNCTION_CALL'),
        #     {},
        # )
        # ```
        self.subworkflow_inputs = arguments  # type:ignore[misc]

        # Call the parent run method to execute the subworkflow
        outputs = {}
        for output in super().run():
            if output.is_fulfilled:
                outputs[output.name] = output.value
            yield output

        # Add the result to the chat history
        self._add_function_result_to_chat_history(outputs, self.state)


class DynamicInlineSubworkflowNode(
    InlineSubworkflowNode[ToolCallingState, BaseInputs, BaseState], FunctionCallNodeMixin
):
    """Node that executes an inline subworkflow with function call output."""

    def run(self) -> Iterator[BaseOutput]:
        arguments = self._extract_function_arguments()

        self.subworkflow_inputs = arguments  # type: ignore[misc]

        # Call the parent run method to execute the subworkflow with proper streaming
        outputs = {}

        for output in super().run():
            if output.is_fulfilled:
                outputs[output.name] = output.value
            yield output

        # Add the result to the chat history
        self._add_function_result_to_chat_history(outputs, self.state)


class FunctionNode(BaseNode[ToolCallingState], FunctionCallNodeMixin):
    """Node that executes a regular Python function with function call output."""

    function_definition: Callable[..., Any]

    class Outputs(BaseNode.Outputs):
        result: Any

    def run(self) -> Iterator[BaseOutput]:
        arguments = self._extract_function_arguments()

        try:
            result = self.function_definition(**arguments)
        except Exception as e:
            self._handle_tool_exception(e, "function", self.function_definition.__name__)

        # Add the result to the chat history
        self._add_function_result_to_chat_history(result, self.state)

        yield BaseOutput(name="result", value=result)


class ComposioNode(BaseNode[ToolCallingState], FunctionCallNodeMixin):
    """Node that executes a Composio tool with function call output."""

    composio_tool: ComposioToolDefinition

    def run(self) -> Iterator[BaseOutput]:
        # Extract arguments from function call
        arguments = self._extract_function_arguments()

        try:
            # Execute using ComposioService
            composio_service = ComposioService()
            if self.composio_tool.user_id is not None:
                result = composio_service.execute_tool(
                    tool_name=self.composio_tool.action, arguments=arguments, user_id=self.composio_tool.user_id
                )
            else:
                result = composio_service.execute_tool(tool_name=self.composio_tool.action, arguments=arguments)
        except Exception as e:
            self._handle_tool_exception(e, "Composio tool", self.composio_tool.action)

        # Add result to chat history
        self._add_function_result_to_chat_history(result, self.state)

        yield from []


class MCPNode(BaseNode[ToolCallingState], FunctionCallNodeMixin):
    """Node that executes an MCP tool with function call output."""

    mcp_tool: MCPToolDefinition

    class Outputs(BaseNode.Outputs):
        result: Any

    def run(self) -> Iterator[BaseOutput]:
        arguments = self._extract_function_arguments()

        try:
            mcp_service = MCPService()
            result = mcp_service.execute_tool(tool_def=self.mcp_tool, arguments=arguments)
        except Exception as e:
            self._handle_tool_exception(e, "MCP tool", self.mcp_tool.name)

        # Add result to chat history
        self._add_function_result_to_chat_history(result, self.state)

        yield BaseOutput(name="result", value=result)


class VellumIntegrationNode(BaseNode[ToolCallingState], FunctionCallNodeMixin):
    """Node that executes a Vellum Integration tool with function call output."""

    vellum_integration_tool: VellumIntegrationToolDefinition

    class Outputs(BaseNode.Outputs):
        result: Any

    def run(self) -> Iterator[BaseOutput]:
        arguments = self._extract_function_arguments()
        vellum_client = self._context.vellum_client

        try:
            vellum_service = VellumIntegrationService(vellum_client)
            result = vellum_service.execute_tool(
                integration=self.vellum_integration_tool.integration_name,
                provider=self.vellum_integration_tool.provider.value,
                tool_name=self.vellum_integration_tool.name,
                arguments=arguments,
            )
        except NodeException as e:
            error_payload = {
                "error": {
                    "code": e.code.value,
                    "message": e.message,
                }
            }
            if e.raw_data is not None:
                error_payload["error"]["raw_data"] = e.raw_data

            self._add_function_result_to_chat_history(error_payload, self.state)
            yield from []
            return
        except Exception as e:
            self._handle_tool_exception(e, "Vellum Integration tool", self.vellum_integration_tool.name)

        # Add result to chat history
        self._add_function_result_to_chat_history(result, self.state)

        yield BaseOutput(name="result", value=result)


class ElseNode(BaseNode[ToolCallingState]):
    """Node that executes when no function conditions match."""

    __exclude_from_monitoring__: bool = True  # Exclude from monitoring views

    class Ports(BaseNode.Ports):
        # Redefined in the create_else_node function, but defined here to resolve mypy errors
        loop_to_router = Port.on_if(
            ToolCallingState.current_prompt_output_index.less_than(ToolPromptNode.Outputs.results.length())
        )
        loop_to_prompt = Port.on_elif(ToolCallingState.current_function_calls_processed.greater_than(0))
        end = Port.on_else()

    def run(self) -> BaseNode.Outputs:
        with self.state.__quiet__():
            self.state.current_prompt_output_index += 1
        return self.Outputs()


def create_tool_prompt_node(
    ml_model: str,
    blocks: List[Union[PromptBlock, Dict[str, Any]]],
    functions: List[Tool],
    prompt_inputs: Optional[EntityInputsInterface],
    parameters: PromptParameters,
    max_prompt_iterations: Optional[int] = None,
    process_parameters_method: Optional[Callable] = None,
    process_blocks_method: Optional[Callable] = None,
    settings: Optional[Union[PromptSettings, Dict[str, Any]]] = None,
) -> Type[ToolPromptNode]:
    if functions and len(functions) > 0:
        prompt_functions: List[Tool] = functions
    else:
        prompt_functions = []

    # Add a chat history block to blocks only if one doesn't already exist
    has_chat_history_block = any(
        (
            (block["block_type"] if isinstance(block, dict) else block.block_type) == "VARIABLE"
            and (
                block["input_variable"]
                if isinstance(block, dict)
                else block.input_variable if isinstance(block, VariablePromptBlock) else None
            )
            == CHAT_HISTORY_VARIABLE
        )
        for block in blocks
    )

    if not has_chat_history_block:
        blocks.append(
            VariablePromptBlock(
                block_type="VARIABLE",
                input_variable=CHAT_HISTORY_VARIABLE,
                state=None,
                cache_config=None,
            )
        )

    node_prompt_inputs = {
        **(prompt_inputs or {}),
        CHAT_HISTORY_VARIABLE: ConcatExpression[List[ChatMessage], List[ChatMessage]](
            lhs=(prompt_inputs or {}).get(CHAT_HISTORY_VARIABLE, []),
            rhs=ToolCallingState.chat_history,
        ),
    }

    # Normalize settings to PromptSettings if provided as a dict
    normalized_settings: Optional[PromptSettings]
    if isinstance(settings, dict):
        normalized_settings = PromptSettings.model_validate(settings)
    else:
        normalized_settings = settings

    node = cast(
        Type[ToolPromptNode],
        type(
            "ToolPromptNode",
            (ToolPromptNode,),
            {
                "ml_model": ml_model,
                "blocks": blocks,
                "functions": prompt_functions,  # Use converted functions for prompt layer
                "prompt_inputs": node_prompt_inputs,
                "parameters": parameters,
                "max_prompt_iterations": max_prompt_iterations,
                "settings": normalized_settings,
                **({"process_parameters": process_parameters_method} if process_parameters_method is not None else {}),
                **({"process_blocks": process_blocks_method} if process_blocks_method is not None else {}),
                "__module__": __name__,
            },
        ),
    )
    return node


def create_router_node(
    functions: List[Tool],
    tool_prompt_node: Type[InlinePromptNode[ToolCallingState]],
) -> Type[RouterNode]:
    """Create a RouterNode with dynamic ports that route based on tool_prompt_node outputs."""

    if functions and len(functions) > 0:
        # Create dynamic ports and convert functions in a single loop
        Ports = type("Ports", (), {})

        # Collect all tool names
        tool_names: List[str] = []
        for function in functions:
            if isinstance(function, MCPServer):
                tool_functions: List[MCPToolDefinition] = compile_mcp_tool_definition(function)
                for tool_function in tool_functions:
                    tool_names.append(get_mcp_tool_name(tool_function))
            else:
                tool_names.append(get_function_name(function))

        # Build conditions for each tool name
        conditions = [
            (
                name,
                ToolCallingState.current_prompt_output_index.less_than(tool_prompt_node.Outputs.results.length())
                & tool_prompt_node.Outputs.results[ToolCallingState.current_prompt_output_index]["type"].equals(
                    "FUNCTION_CALL"
                )
                & tool_prompt_node.Outputs.results[ToolCallingState.current_prompt_output_index]["value"][
                    "name"
                ].equals(name),
            )
            for name in tool_names
        ]

        # Assign ports: first condition uses on_if, subsequent ones use on_elif
        for idx, (name, condition) in enumerate(conditions):
            port = Port.on_if(condition) if idx == 0 else Port.on_elif(condition)
            setattr(Ports, name, port)

        # Add the else port for when no function conditions match
        setattr(Ports, "default", Port.on_else())
    else:
        # If no functions exist, create a simple Ports class with just a default port
        Ports = type("Ports", (), {"default": Port(default=True)})

    node = cast(
        Type[RouterNode],
        type(
            "RouterNode",
            (RouterNode,),
            {
                "Ports": Ports,
                "prompt_outputs": tool_prompt_node.Outputs.results,
                "__module__": __name__,
            },
        ),
    )
    return node


def create_function_node(
    function: ToolBase,
    tool_prompt_node: Type[ToolPromptNode],
) -> Type[BaseNode]:
    """
    Create a FunctionNode class for a given function.

    For workflow functions: BaseNode
    For regular functions: BaseNode with direct function call

    Args:
        function: The function to create a node for
        tool_prompt_node: The tool prompt node class
    """
    if isinstance(function, DeploymentDefinition):
        deployment = function.deployment_id or function.deployment_name
        release_tag = function.release_tag

        node = type(
            f"DynamicSubworkflowDeploymentNode_{deployment}",
            (DynamicSubworkflowDeploymentNode,),
            {
                "deployment": deployment,
                "release_tag": release_tag,
                "function_call_output": tool_prompt_node.Outputs.results,
                "__module__": __name__,
            },
        )

        return node

    elif isinstance(function, ComposioToolDefinition):
        node = type(
            f"ComposioNode_{function.name}",
            (ComposioNode,),
            {
                "composio_tool": function,
                "function_call_output": tool_prompt_node.Outputs.results,
                "__module__": __name__,
            },
        )
        return node
    elif isinstance(function, VellumIntegrationToolDefinition):
        node = type(
            f"VellumIntegrationNode_{function.name}",
            (VellumIntegrationNode,),
            {
                "vellum_integration_tool": function,
                "function_call_output": tool_prompt_node.Outputs.results,
                "__module__": __name__,
            },
        )
        return node
    elif is_workflow_class(function):
        function.is_dynamic = True
        node = type(
            f"DynamicInlineSubworkflowNode_{function.__name__}",
            (DynamicInlineSubworkflowNode,),
            {
                "subworkflow": function,
                "function_call_output": tool_prompt_node.Outputs.results,
                "__module__": __name__,
            },
        )
    else:

        def create_function_wrapper(func):
            def wrapper(self, **kwargs):
                merged_kwargs = kwargs.copy()
                inputs = getattr(func, "__vellum_inputs__", {})
                if inputs:
                    for param_name, param_ref in inputs.items():
                        if isinstance(param_ref, BaseDescriptor):
                            resolved_value = param_ref.resolve(self.state)
                        else:
                            resolved_value = param_ref
                        merged_kwargs[param_name] = resolved_value

                return func(**merged_kwargs)

            # Preserve the original function name for better error messages
            wrapper.__name__ = func.__name__
            return wrapper

        node = type(
            f"FunctionNode_{function.__name__}",
            (FunctionNode,),
            {
                "function_definition": create_function_wrapper(function),
                "function_call_output": tool_prompt_node.Outputs.results,
                "__module__": __name__,
            },
        )

    return node


def create_mcp_tool_node(
    tool_def: MCPToolDefinition,
    tool_prompt_node: Type[ToolPromptNode],
) -> Type[BaseNode]:
    node = type(
        f"MCPNode_{tool_def.name}",
        (MCPNode,),
        {
            "mcp_tool": tool_def,
            "function_call_output": tool_prompt_node.Outputs.results,
            "__module__": __name__,
        },
    )
    return node


def create_else_node(
    tool_prompt_node: Type[ToolPromptNode],
) -> Type[ElseNode]:
    class Ports(ElseNode.Ports):
        loop_to_router = Port.on_if(
            ToolCallingState.current_prompt_output_index.less_than(tool_prompt_node.Outputs.results.length())
        )
        loop_to_prompt = Port.on_elif(ToolCallingState.current_function_calls_processed.greater_than(0))
        end = Port.on_else()

    node = cast(
        Type[ElseNode],
        type(
            f"{tool_prompt_node.__name__}_ElseNode",
            (ElseNode,),
            {
                "Ports": Ports,
                "__module__": __name__,
            },
        ),
    )
    return node


def get_function_name(function: ToolBase) -> str:
    if isinstance(function, DeploymentDefinition):
        name = str(function.deployment_id or function.deployment_name)
        return name.replace("-", "")
    elif isinstance(function, ComposioToolDefinition):
        # model post init sets the name to the action if it's not set
        return function.name  # type: ignore[return-value]
    elif isinstance(function, VellumIntegrationToolDefinition):
        return function.name
    else:
        return snake_case(function.__name__)



from typing import Generic, TypeVar

from vellum.workflows.nodes.displayable.tool_calling_node.utils import FunctionNode
from vellum_ee.workflows.display.editor.types import NodeDisplayData
from vellum_ee.workflows.display.nodes.base_node_display import BaseNodeDisplay

_FunctionNodeType = TypeVar("_FunctionNodeType", bound=FunctionNode)


class FunctionNodeDisplay(BaseNodeDisplay[_FunctionNodeType], Generic[_FunctionNodeType]):
    display_data = NodeDisplayData(
        icon="vellum:icon:rectangle-code",
        color="purple",
    )



# This file was auto-generated by Fern from our API Definition.

import typing

ApiActorTypeEnum = typing.Union[
    typing.Literal["WORKSPACE_API_KEY", "ENVIRONMENT_API_KEY", "JWT", "SERVICE_TOKEN"], typing.Any
]



# This file was auto-generated by Fern from our API Definition.

import typing

import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel
from .deployment_release_tag_deployment_history_item import DeploymentReleaseTagDeploymentHistoryItem
from .release_tag_release import ReleaseTagRelease
from .release_tag_source import ReleaseTagSource


class DeploymentReleaseTagRead(UniversalBaseModel):
    name: str = pydantic.Field()
    """
    The name of the Release Tag
    """

    source: ReleaseTagSource = pydantic.Field()
    """
    The source of how the Release Tag was originally created
    
    * `SYSTEM` - System
    * `USER` - User
    """

    history_item: DeploymentReleaseTagDeploymentHistoryItem = pydantic.Field()
    """
    Deprecated. Reference the `release` field instead.
    """

    release: ReleaseTagRelease = pydantic.Field()
    """
    The Release that this Release Tag points to.
    """

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



"""Monitoring execution context for workflow tracing."""

import threading
from uuid import UUID
from typing import TYPE_CHECKING, Dict, Optional

if TYPE_CHECKING:
    from vellum.workflows.context import ExecutionContext

DEFAULT_TRACE_ID = UUID("00000000-0000-0000-0000-000000000000")


class MonitoringContextStore:
    """
    thread-safe storage for monitoring contexts.
    handles context persistence and retrieval across threads.
    """

    def __init__(self):
        self._lock = threading.Lock()
        self._contexts: Dict[str, "ExecutionContext"] = {}

    def get_current_trace_id(self) -> Optional[UUID]:
        """Get the current active trace_id that should be used by all threads."""
        with self._lock:
            current_context = self.retrieve_context()
            return current_context.trace_id if current_context else None

    def store_context(self, context: "ExecutionContext") -> None:
        """Store monitoring parent context using trace:span:thread keys."""
        thread_id = threading.get_ident()
        current_thread = threading.current_thread()
        trace_id = context.trace_id

        with self._lock:
            # Get span_id from RelationalThread first
            span_id = None

            if context.parent_context and hasattr(context.parent_context, "span_id"):
                span_id = context.parent_context.span_id

            if (
                not span_id
                and hasattr(current_thread, "get_parent_span_id")
                and callable(current_thread.get_parent_span_id)
            ):
                span_id = current_thread.get_parent_span_id()

            # Always use trace:span:thread format - require span_id
            if span_id:
                context_key = f"trace:{str(trace_id)}:span:{str(span_id)}:thread:{thread_id}"
                self._contexts[context_key] = context

    def retrieve_context(self) -> Optional["ExecutionContext"]:
        """Retrieve monitoring parent context using trace:span:thread keys."""
        current_thread = threading.current_thread()
        current_thread_id = current_thread.ident

        # Get trace_id and span_id directly from the thread if it's a RelationalThread
        trace_id = None
        span_id = None
        if hasattr(current_thread, "get_trace_id") and callable(current_thread.get_trace_id):
            trace_id = current_thread.get_trace_id()
        if hasattr(current_thread, "get_parent_span_id") and callable(current_thread.get_parent_span_id):
            span_id = current_thread.get_parent_span_id()

        # Require both trace_id and span_id - no fallback searching
        if not trace_id or not span_id:
            return None

        with self._lock:
            # Try current thread
            current_key = f"trace:{str(trace_id)}:span:{str(span_id)}:thread:{current_thread_id}"
            if current_key in self._contexts:
                return self._contexts[current_key]

            # Try parent thread with same trace and span
            if hasattr(current_thread, "get_parent_thread") and callable(current_thread.get_parent_thread):
                parent_thread_id = current_thread.get_parent_thread()
                if parent_thread_id:
                    parent_key = f"trace:{str(trace_id)}:span:{str(span_id)}:thread:{parent_thread_id}"
                    if parent_key in self._contexts:
                        return self._contexts[parent_key]

        return None

    def clear_context(self):
        """Clear all stored contexts."""
        with self._lock:
            self._contexts.clear()



import pytest
import base64
from unittest.mock import Mock, patch

from vellum import VellumAudio, VellumDocument, VellumImage, VellumVideo
from vellum.utils.files import InvalidFileSourceError, stream_vellum_file

# Sample content for different file types
SAMPLE_TEXT_CONTENT = b"This is a sample text document content."
SAMPLE_IMAGE_CONTENT = b"\x89PNG\r\n\x1a\n\x00\x00\x00\rIHDR\x00\x00\x00\x01"  # Minimal PNG header
SAMPLE_AUDIO_CONTENT = b"RIFF\x00\x00\x00\x00WAVEfmt "  # Minimal WAV header
SAMPLE_VIDEO_CONTENT = b"\x00\x00\x00\x20ftypmp42"  # Minimal MP4 header


@pytest.mark.parametrize(
    ["file_type", "content", "mime_type"],
    [
        (VellumDocument, SAMPLE_TEXT_CONTENT, "text/plain"),
        (VellumImage, SAMPLE_IMAGE_CONTENT, "image/png"),
        (VellumAudio, SAMPLE_AUDIO_CONTENT, "audio/wav"),
        (VellumVideo, SAMPLE_VIDEO_CONTENT, "video/mp4"),
    ],
)
def test_stream_vellum_file_base64_data_url(file_type, content, mime_type):
    """Test streaming files from base64 data URLs."""
    # Encode content as base64
    base64_content = base64.b64encode(content).decode("utf-8")
    src = f"data:{mime_type};base64,{base64_content}"

    # Create the appropriate file type
    vellum_file = file_type(src=src)

    # Stream the file and collect chunks
    chunks = []
    with stream_vellum_file(vellum_file) as chunk_iter:
        for chunk in chunk_iter:
            chunks.append(chunk)

    # Verify the content matches
    result = b"".join(chunks)
    assert result == content


@pytest.mark.parametrize(
    ["file_type", "content"],
    [
        (VellumDocument, SAMPLE_TEXT_CONTENT),
        (VellumImage, SAMPLE_IMAGE_CONTENT),
        (VellumAudio, SAMPLE_AUDIO_CONTENT),
        (VellumVideo, SAMPLE_VIDEO_CONTENT),
    ],
)
def test_stream_vellum_file_with_custom_chunk_size(file_type, content):
    """Test that custom chunk size is respected."""
    # Encode content as base64
    base64_content = base64.b64encode(content).decode("utf-8")
    src = f"data:application/octet-stream;base64,{base64_content}"

    vellum_file = file_type(src=src)

    # Stream with small chunk size
    chunk_size = 5
    chunks = []
    with stream_vellum_file(vellum_file, chunk_size=chunk_size) as chunk_iter:
        for chunk in chunk_iter:
            chunks.append(chunk)
            # Each chunk should be <= chunk_size (last chunk may be smaller)
            assert len(chunk) <= chunk_size

    # Verify the content matches when reassembled
    result = b"".join(chunks)
    assert result == content

    # Verify we got multiple chunks for content larger than chunk_size
    if len(content) > chunk_size:
        assert len(chunks) > 1


@pytest.mark.parametrize(
    "file_type",
    [VellumDocument, VellumImage, VellumAudio, VellumVideo],
)
def test_stream_vellum_file_from_url(file_type):
    """Test streaming files from direct URLs."""
    content = b"URL fetched content that will be streamed"
    url = "https://example.com/file.dat"

    vellum_file = file_type(src=url)

    with patch("vellum.utils.files.stream.requests.get") as mock_get:
        mock_response = Mock()
        # Simulate streaming response with iter_content
        mock_response.iter_content = Mock(return_value=iter([content[:20], content[20:]]))
        mock_response.raise_for_status = Mock()
        mock_response.close = Mock()
        mock_get.return_value = mock_response

        chunks = []
        with stream_vellum_file(vellum_file) as chunk_iter:
            for chunk in chunk_iter:
                chunks.append(chunk)

        result = b"".join(chunks)
        assert result == content
        mock_get.assert_called_once_with(url, stream=True)
        mock_response.raise_for_status.assert_called_once()
        mock_response.close.assert_called_once()


@pytest.mark.parametrize(
    "file_type",
    [VellumDocument, VellumImage, VellumAudio, VellumVideo],
)
def test_stream_vellum_file_from_vellum_uploaded_file(file_type):
    """Test streaming files from Vellum uploaded file IDs."""
    content = b"Vellum uploaded content to stream"
    file_id = "12345678-1234-1234-1234-123456789abc"
    src = f"vellum:uploaded-file:{file_id}"
    file_url = "https://s3.amazonaws.com/example/file.dat"

    vellum_file = file_type(src=src)

    with patch("vellum.utils.vellum_client.create_vellum_client") as mock_create_client, patch(
        "vellum.utils.files.stream.requests.get"
    ) as mock_get:
        # Mock the Vellum client
        mock_client = Mock()
        uploaded_file_mock = Mock()
        uploaded_file_mock.file_url = file_url
        mock_client.uploaded_files.retrieve.return_value = uploaded_file_mock
        mock_create_client.return_value = mock_client

        # Mock the file download with streaming
        file_response = Mock()
        file_response.iter_content = Mock(return_value=iter([content[:15], content[15:]]))
        file_response.raise_for_status = Mock()
        file_response.close = Mock()

        mock_get.return_value = file_response

        chunks = []
        with stream_vellum_file(vellum_file) as chunk_iter:
            for chunk in chunk_iter:
                chunks.append(chunk)

        result = b"".join(chunks)
        assert result == content
        mock_client.uploaded_files.retrieve.assert_called_once_with(file_id)
        mock_get.assert_called_once_with(file_url, stream=True)
        file_response.close.assert_called_once()


def test_stream_vellum_file_context_manager_cleanup():
    """Test that the context manager properly cleans up resources."""
    content = b"Test content"
    base64_content = base64.b64encode(content).decode("utf-8")
    src = f"data:text/plain;base64,{base64_content}"

    vellum_file = VellumDocument(src=src)

    # Test that we can use the iterator only within the context
    with stream_vellum_file(vellum_file) as chunk_iter:
        chunks = list(chunk_iter)

    # After the context, the iterator should be exhausted
    assert b"".join(chunks) == content


def test_stream_vellum_file_invalid_src():
    """Test that invalid src raises InvalidFileSourceError even in streaming mode."""
    vellum_file = VellumDocument(src="invalid://source")

    with pytest.raises(InvalidFileSourceError, match="Invalid file source"):
        with stream_vellum_file(vellum_file) as chunk_iter:
            list(chunk_iter)


def test_stream_vellum_file_large_file_simulation():
    """Test streaming a large file to verify memory efficiency."""
    # Simulate a large file by creating content in chunks
    large_content = b"x" * 10000  # 10KB
    base64_content = base64.b64encode(large_content).decode("utf-8")
    src = f"data:application/octet-stream;base64,{base64_content}"

    vellum_file = VellumDocument(src=src)

    # Stream with 1KB chunks
    chunk_size = 1024
    total_bytes = 0
    chunk_count = 0

    with stream_vellum_file(vellum_file, chunk_size=chunk_size) as chunk_iter:
        for chunk in chunk_iter:
            total_bytes += len(chunk)
            chunk_count += 1
            assert len(chunk) <= chunk_size

    assert total_bytes == len(large_content)
    assert chunk_count >= len(large_content) // chunk_size



import pytest
import base64
from unittest.mock import Mock, patch

from vellum import VellumAudio, VellumDocument, VellumImage, VellumVideo
from vellum.utils.files import FileRetrievalError, InvalidFileSourceError, read_vellum_file

# Sample content for different file types
SAMPLE_TEXT_CONTENT = b"This is a sample text document content."
SAMPLE_IMAGE_CONTENT = b"\x89PNG\r\n\x1a\n\x00\x00\x00\rIHDR\x00\x00\x00\x01"  # Minimal PNG header
SAMPLE_AUDIO_CONTENT = b"RIFF\x00\x00\x00\x00WAVEfmt "  # Minimal WAV header
SAMPLE_VIDEO_CONTENT = b"\x00\x00\x00\x20ftypmp42"  # Minimal MP4 header


@pytest.mark.parametrize(
    ["file_type", "content", "mime_type"],
    [
        (VellumDocument, SAMPLE_TEXT_CONTENT, "text/plain"),
        (VellumImage, SAMPLE_IMAGE_CONTENT, "image/png"),
        (VellumAudio, SAMPLE_AUDIO_CONTENT, "audio/wav"),
        (VellumVideo, SAMPLE_VIDEO_CONTENT, "video/mp4"),
    ],
)
def test_read_vellum_file_base64_data_url_with_mime_type(file_type, content, mime_type):
    """Test reading files from base64 data URLs with MIME types."""
    # Encode content as base64
    base64_content = base64.b64encode(content).decode("utf-8")
    src = f"data:{mime_type};base64,{base64_content}"

    # Create the appropriate file type
    vellum_file = file_type(src=src)

    # Read the file
    result = read_vellum_file(vellum_file)

    # Verify the content matches
    assert result == content


@pytest.mark.parametrize(
    ["file_type", "content"],
    [
        (VellumDocument, SAMPLE_TEXT_CONTENT),
        (VellumImage, SAMPLE_IMAGE_CONTENT),
        (VellumAudio, SAMPLE_AUDIO_CONTENT),
        (VellumVideo, SAMPLE_VIDEO_CONTENT),
    ],
)
def test_read_vellum_file_base64_data_url_without_mime_type(file_type, content):
    """Test reading files from base64 data URLs without MIME types."""
    # Encode content as base64
    base64_content = base64.b64encode(content).decode("utf-8")
    src = f"data:;base64,{base64_content}"

    # Create the appropriate file type
    vellum_file = file_type(src=src)

    # Read the file
    result = read_vellum_file(vellum_file)

    # Verify the content matches
    assert result == content


@pytest.mark.parametrize(
    ["file_type", "content", "mime_type", "charset"],
    [
        (VellumDocument, SAMPLE_TEXT_CONTENT, "text/plain", "utf-8"),
        (VellumDocument, SAMPLE_TEXT_CONTENT, "text/html", "iso-8859-1"),
        (VellumImage, SAMPLE_IMAGE_CONTENT, "image/png", "utf-8"),
    ],
)
def test_read_vellum_file_base64_data_url_with_charset(file_type, content, mime_type, charset):
    """Test reading files from base64 data URLs with charset specifications."""
    # Encode content as base64
    base64_content = base64.b64encode(content).decode("utf-8")
    src = f"data:{mime_type};charset={charset};base64,{base64_content}"

    # Create the appropriate file type
    vellum_file = file_type(src=src)

    # Read the file
    result = read_vellum_file(vellum_file)

    # Verify the content matches
    assert result == content


@pytest.mark.parametrize(
    ["file_type", "content"],
    [
        (VellumDocument, b""),  # Empty file
        (VellumImage, b"x" * 1000),  # Larger content
        (VellumAudio, b"\x00\x01\x02\x03\x04"),  # Binary content with special bytes
        (VellumVideo, b"Unicode content: \xc3\xa9\xc3\xa0\xc3\xb1"),  # UTF-8 encoded content
    ],
)
def test_read_vellum_file_base64_edge_cases(file_type, content):
    """Test reading files with edge case content (empty, large, special characters)."""
    # Encode content as base64
    base64_content = base64.b64encode(content).decode("utf-8")
    src = f"data:application/octet-stream;base64,{base64_content}"

    # Create the appropriate file type
    vellum_file = file_type(src=src)

    # Read the file
    result = read_vellum_file(vellum_file)

    # Verify the content matches
    assert result == content


@pytest.mark.parametrize(
    "file_type",
    [VellumDocument, VellumImage, VellumAudio, VellumVideo],
)
def test_read_vellum_file_from_url(file_type):
    """Test reading files from direct URLs."""
    content = b"URL fetched content"
    url = "https://example.com/file.dat"

    vellum_file = file_type(src=url)

    with patch("vellum.utils.files.stream.requests.get") as mock_get:
        mock_response = Mock()
        # Mock streaming interface
        mock_response.iter_content = Mock(return_value=iter([content]))
        mock_response.raise_for_status = Mock()
        mock_response.close = Mock()
        mock_get.return_value = mock_response

        result = read_vellum_file(vellum_file)

        assert result == content
        mock_get.assert_called_once_with(url, stream=True)
        mock_response.raise_for_status.assert_called_once()
        mock_response.close.assert_called_once()


@pytest.mark.parametrize(
    "file_type",
    [VellumDocument, VellumImage, VellumAudio, VellumVideo],
)
def test_read_vellum_file_from_vellum_uploaded_file(file_type):
    """Test reading files from Vellum uploaded file IDs."""
    content = b"Vellum uploaded content"
    file_id = "12345678-1234-1234-1234-123456789abc"
    src = f"vellum:uploaded-file:{file_id}"
    file_url = "https://s3.amazonaws.com/example/file.dat"

    vellum_file = file_type(src=src)

    with patch("vellum.utils.vellum_client.create_vellum_client") as mock_create_client, patch(
        "vellum.utils.files.stream.requests.get"
    ) as mock_get:
        # Mock the Vellum client
        mock_client = Mock()
        uploaded_file_mock = Mock()
        uploaded_file_mock.file_url = file_url
        mock_client.uploaded_files.retrieve.return_value = uploaded_file_mock
        mock_create_client.return_value = mock_client

        # Mock the file download with streaming interface
        file_response = Mock()
        file_response.iter_content = Mock(return_value=iter([content]))
        file_response.raise_for_status = Mock()
        file_response.close = Mock()

        mock_get.return_value = file_response

        result = read_vellum_file(vellum_file)

        assert result == content
        mock_client.uploaded_files.retrieve.assert_called_once_with(file_id)
        mock_get.assert_called_once_with(file_url, stream=True)
        file_response.close.assert_called_once()


def test_read_vellum_file_invalid_src():
    """Test that invalid src raises InvalidFileSourceError."""
    vellum_file = VellumDocument(src="invalid://source")

    with pytest.raises(InvalidFileSourceError, match="Invalid file source"):
        read_vellum_file(vellum_file)


def test_read_vellum_file_vellum_uploaded_file_missing_url():
    """Test that missing file_url in API response raises FileRetrievalError."""
    file_id = "12345678-1234-1234-1234-123456789abc"
    src = f"vellum:uploaded-file:{file_id}"

    vellum_file = VellumDocument(src=src)

    # Mock the create_vellum_client to return a client with an uploaded file without a file_url
    with patch("vellum.utils.vellum_client.create_vellum_client") as mock_create_client:
        mock_client = Mock()
        uploaded_file_mock = Mock()
        uploaded_file_mock.file_url = None
        mock_client.uploaded_files.retrieve.return_value = uploaded_file_mock
        mock_create_client.return_value = mock_client

        with pytest.raises(FileRetrievalError, match="has no accessible URL"):
            read_vellum_file(vellum_file)



# This file was auto-generated by Fern from our API Definition.

import typing
from json.decoder import JSONDecodeError

from ...core.api_error import ApiError
from ...core.client_wrapper import AsyncClientWrapper, SyncClientWrapper
from ...core.http_response import AsyncHttpResponse, HttpResponse
from ...core.jsonable_encoder import jsonable_encoder
from ...core.pydantic_utilities import parse_obj_as
from ...core.request_options import RequestOptions
from ...core.serialization import convert_and_respect_annotation_metadata
from ...types.metric_definition_execution import MetricDefinitionExecution
from ...types.metric_definition_history_item import MetricDefinitionHistoryItem
from ...types.metric_definition_input import MetricDefinitionInput

# this is used as the default value for optional parameters
OMIT = typing.cast(typing.Any, ...)


class RawMetricDefinitionsClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._client_wrapper = client_wrapper

    def execute_metric_definition(
        self,
        id: str,
        *,
        inputs: typing.Sequence[MetricDefinitionInput],
        release_tag: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[MetricDefinitionExecution]:
        """
        Parameters
        ----------
        id : str
            Either the Metric Definition's ID or its unique name

        inputs : typing.Sequence[MetricDefinitionInput]

        release_tag : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[MetricDefinitionExecution]

        """
        _response = self._client_wrapper.httpx_client.request(
            f"v1/metric-definitions/{jsonable_encoder(id)}/execute",
            base_url=self._client_wrapper.get_environment().predict,
            method="POST",
            json={
                "inputs": convert_and_respect_annotation_metadata(
                    object_=inputs, annotation=typing.Sequence[MetricDefinitionInput], direction="write"
                ),
                "release_tag": release_tag,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    MetricDefinitionExecution,
                    parse_obj_as(
                        type_=MetricDefinitionExecution,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def metric_definition_history_item_retrieve(
        self, history_id_or_release_tag: str, id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[MetricDefinitionHistoryItem]:
        """
        Parameters
        ----------
        history_id_or_release_tag : str
            Either the UUID of Metric Definition History Item you'd like to retrieve, or the name of a Release Tag that's pointing to the Metric Definition History Item you'd like to retrieve.

        id : str
            A UUID string identifying this metric definition.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[MetricDefinitionHistoryItem]

        """
        _response = self._client_wrapper.httpx_client.request(
            f"v1/metric-definitions/{jsonable_encoder(id)}/history/{jsonable_encoder(history_id_or_release_tag)}",
            base_url=self._client_wrapper.get_environment().default,
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    MetricDefinitionHistoryItem,
                    parse_obj_as(
                        type_=MetricDefinitionHistoryItem,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)


class AsyncRawMetricDefinitionsClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._client_wrapper = client_wrapper

    async def execute_metric_definition(
        self,
        id: str,
        *,
        inputs: typing.Sequence[MetricDefinitionInput],
        release_tag: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[MetricDefinitionExecution]:
        """
        Parameters
        ----------
        id : str
            Either the Metric Definition's ID or its unique name

        inputs : typing.Sequence[MetricDefinitionInput]

        release_tag : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[MetricDefinitionExecution]

        """
        _response = await self._client_wrapper.httpx_client.request(
            f"v1/metric-definitions/{jsonable_encoder(id)}/execute",
            base_url=self._client_wrapper.get_environment().predict,
            method="POST",
            json={
                "inputs": convert_and_respect_annotation_metadata(
                    object_=inputs, annotation=typing.Sequence[MetricDefinitionInput], direction="write"
                ),
                "release_tag": release_tag,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    MetricDefinitionExecution,
                    parse_obj_as(
                        type_=MetricDefinitionExecution,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def metric_definition_history_item_retrieve(
        self, history_id_or_release_tag: str, id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[MetricDefinitionHistoryItem]:
        """
        Parameters
        ----------
        history_id_or_release_tag : str
            Either the UUID of Metric Definition History Item you'd like to retrieve, or the name of a Release Tag that's pointing to the Metric Definition History Item you'd like to retrieve.

        id : str
            A UUID string identifying this metric definition.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[MetricDefinitionHistoryItem]

        """
        _response = await self._client_wrapper.httpx_client.request(
            f"v1/metric-definitions/{jsonable_encoder(id)}/history/{jsonable_encoder(history_id_or_release_tag)}",
            base_url=self._client_wrapper.get_environment().default,
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    MetricDefinitionHistoryItem,
                    parse_obj_as(
                        type_=MetricDefinitionHistoryItem,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)



# This file was auto-generated by Fern from our API Definition.

import typing

import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel
from .search_result_document import SearchResultDocument
from .search_result_meta import SearchResultMeta


class SearchResult(UniversalBaseModel):
    text: str = pydantic.Field()
    """
    The text of the chunk that matched the search query.
    """

    score: float = pydantic.Field()
    """
    A score representing how well the chunk matches the search query.
    """

    keywords: typing.List[str]
    document: SearchResultDocument = pydantic.Field()
    """
    The document that contains the chunk that matched the search query.
    """

    meta: typing.Optional[SearchResultMeta] = pydantic.Field(default=None)
    """
    Additional information about the search result.
    """

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



class UserFacingException(Exception):
    def to_message(self) -> str:
        return str(self)


class UnsupportedSerializationException(UserFacingException):
    pass


class NodeValidationError(UserFacingException):
    """Exception raised when a node fails validation during workflow display serialization."""

    def __init__(self, message: str, node_class_name: str):
        self.message = message
        self.node_class_name = node_class_name
        super().__init__(f"Node validation error in {node_class_name}: {message}")


class InvalidInputReferenceError(UserFacingException):
    """Exception raised when a node references a non-existent workflow input."""

    def __init__(self, message: str, inputs_class_name: str, attribute_name: str):
        self.message = message
        self.inputs_class_name = inputs_class_name
        self.attribute_name = attribute_name
        super().__init__(f"Invalid input reference in {inputs_class_name}.{attribute_name}: {message}")



# This file was auto-generated by Fern from our API Definition.

import typing

from ...core.client_wrapper import AsyncClientWrapper, SyncClientWrapper
from ...core.request_options import RequestOptions
from ...types.container_image_read import ContainerImageRead
from ...types.docker_service_token import DockerServiceToken
from ...types.paginated_container_image_read_list import PaginatedContainerImageReadList
from .raw_client import AsyncRawContainerImagesClient, RawContainerImagesClient

# this is used as the default value for optional parameters
OMIT = typing.cast(typing.Any, ...)


class ContainerImagesClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._raw_client = RawContainerImagesClient(client_wrapper=client_wrapper)

    @property
    def _client_wrapper(self) -> SyncClientWrapper:
        return self._raw_client._client_wrapper

    @property
    def with_raw_response(self) -> RawContainerImagesClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        RawContainerImagesClient
        """
        return self._raw_client

    def list(
        self,
        *,
        limit: typing.Optional[int] = None,
        offset: typing.Optional[int] = None,
        ordering: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> PaginatedContainerImageReadList:
        """
        Retrieve a list of container images for the organization.

        Parameters
        ----------
        limit : typing.Optional[int]
            Number of results to return per page.

        offset : typing.Optional[int]
            The initial index from which to return the results.

        ordering : typing.Optional[str]
            Which field to use when ordering the results.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        PaginatedContainerImageReadList


        Examples
        --------
        from vellum import Vellum

        client = Vellum(
            api_version="YOUR_API_VERSION",
            api_key="YOUR_API_KEY",
        )
        client.container_images.list()
        """
        _response = self._raw_client.list(
            limit=limit, offset=offset, ordering=ordering, request_options=request_options
        )
        return _response.data

    def retrieve(self, id: str, *, request_options: typing.Optional[RequestOptions] = None) -> ContainerImageRead:
        """
        Retrieve a Container Image by its ID or name.

        Parameters
        ----------
        id : str
            Either the Container Image's ID or its unique name

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        ContainerImageRead


        Examples
        --------
        from vellum import Vellum

        client = Vellum(
            api_version="YOUR_API_VERSION",
            api_key="YOUR_API_KEY",
        )
        client.container_images.retrieve(
            id="id",
        )
        """
        _response = self._raw_client.retrieve(id, request_options=request_options)
        return _response.data

    def docker_service_token(self, *, request_options: typing.Optional[RequestOptions] = None) -> DockerServiceToken:
        """
        Parameters
        ----------
        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        DockerServiceToken


        Examples
        --------
        from vellum import Vellum

        client = Vellum(
            api_version="YOUR_API_VERSION",
            api_key="YOUR_API_KEY",
        )
        client.container_images.docker_service_token()
        """
        _response = self._raw_client.docker_service_token(request_options=request_options)
        return _response.data

    def push_container_image(
        self,
        *,
        name: str,
        sha: str,
        tags: typing.Sequence[str],
        force: typing.Optional[bool] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> ContainerImageRead:
        """
        Parameters
        ----------
        name : str

        sha : str

        tags : typing.Sequence[str]

        force : typing.Optional[bool]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        ContainerImageRead


        Examples
        --------
        from vellum import Vellum

        client = Vellum(
            api_version="YOUR_API_VERSION",
            api_key="YOUR_API_KEY",
        )
        client.container_images.push_container_image(
            name="x",
            sha="x",
            tags=["tags", "tags"],
        )
        """
        _response = self._raw_client.push_container_image(
            name=name, sha=sha, tags=tags, force=force, request_options=request_options
        )
        return _response.data


class AsyncContainerImagesClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._raw_client = AsyncRawContainerImagesClient(client_wrapper=client_wrapper)

    @property
    def with_raw_response(self) -> AsyncRawContainerImagesClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        AsyncRawContainerImagesClient
        """
        return self._raw_client

    async def list(
        self,
        *,
        limit: typing.Optional[int] = None,
        offset: typing.Optional[int] = None,
        ordering: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> PaginatedContainerImageReadList:
        """
        Retrieve a list of container images for the organization.

        Parameters
        ----------
        limit : typing.Optional[int]
            Number of results to return per page.

        offset : typing.Optional[int]
            The initial index from which to return the results.

        ordering : typing.Optional[str]
            Which field to use when ordering the results.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        PaginatedContainerImageReadList


        Examples
        --------
        import asyncio

        from vellum import AsyncVellum

        client = AsyncVellum(
            api_version="YOUR_API_VERSION",
            api_key="YOUR_API_KEY",
        )


        async def main() -> None:
            await client.container_images.list()


        asyncio.run(main())
        """
        _response = await self._raw_client.list(
            limit=limit, offset=offset, ordering=ordering, request_options=request_options
        )
        return _response.data

    async def retrieve(self, id: str, *, request_options: typing.Optional[RequestOptions] = None) -> ContainerImageRead:
        """
        Retrieve a Container Image by its ID or name.

        Parameters
        ----------
        id : str
            Either the Container Image's ID or its unique name

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        ContainerImageRead


        Examples
        --------
        import asyncio

        from vellum import AsyncVellum

        client = AsyncVellum(
            api_version="YOUR_API_VERSION",
            api_key="YOUR_API_KEY",
        )


        async def main() -> None:
            await client.container_images.retrieve(
                id="id",
            )


        asyncio.run(main())
        """
        _response = await self._raw_client.retrieve(id, request_options=request_options)
        return _response.data

    async def docker_service_token(
        self, *, request_options: typing.Optional[RequestOptions] = None
    ) -> DockerServiceToken:
        """
        Parameters
        ----------
        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        DockerServiceToken


        Examples
        --------
        import asyncio

        from vellum import AsyncVellum

        client = AsyncVellum(
            api_version="YOUR_API_VERSION",
            api_key="YOUR_API_KEY",
        )


        async def main() -> None:
            await client.container_images.docker_service_token()


        asyncio.run(main())
        """
        _response = await self._raw_client.docker_service_token(request_options=request_options)
        return _response.data

    async def push_container_image(
        self,
        *,
        name: str,
        sha: str,
        tags: typing.Sequence[str],
        force: typing.Optional[bool] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> ContainerImageRead:
        """
        Parameters
        ----------
        name : str

        sha : str

        tags : typing.Sequence[str]

        force : typing.Optional[bool]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        ContainerImageRead


        Examples
        --------
        import asyncio

        from vellum import AsyncVellum

        client = AsyncVellum(
            api_version="YOUR_API_VERSION",
            api_key="YOUR_API_KEY",
        )


        async def main() -> None:
            await client.container_images.push_container_image(
                name="x",
                sha="x",
                tags=["tags", "tags"],
            )


        asyncio.run(main())
        """
        _response = await self._raw_client.push_container_image(
            name=name, sha=sha, tags=tags, force=force, request_options=request_options
        )
        return _response.data
